// -*- C++ -*-
/*!
 * @file  PositionController.cpp * @brief Positional output  Controller * $Date$ 
 *
 * $Id$ 
 */
#include "PositionController.h"
#include "VectorConvert.h"


// Module specification
static const char* positioncontroller_spec[] =
  {
    "implementation_id", "UserController",
    "type_name",         "UserController",
    "description",       "Position commander",
    "version",           "1.0",
    "vendor",            "Shinji Kurihara",
    "category",          "Controller",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.diag_Pgain", "1.0,2.0,3.0,4.0,5.0,6.0",
    "conf.default.diag_Dgain", "2.0,4.0,6.0,8.0,10.0,12.0",
    ""
  };

PositionController::PositionController(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_PositionInput("in0", m_in0),
    m_VelocityInput("in1", m_in1),
    m_PositionOutput("TargetValue", m_TargetValue),
    m_ConfigurationPort("Configuration")
{
  // Set InPort buffers
  registerInPort("Current Position", m_PositionInput);
  registerInPort("Current Velocity", m_VelocityInput);
  
  // Set OutPort buffer
  registerOutPort("Target Position", m_PositionOutput);
  
  // Set service consumers to Ports
  m_ConfigurationPort.registerConsumer("PDControlSvc0", "PDControlService", m_PDControlSvc0);

  // Set CORBA Service Ports
  registerPort(m_ConfigurationPort);
}

PositionController::~PositionController()
{
}


RTC::ReturnCode_t PositionController::onInitialize()
{
  // Bind variables and configuration variable
  bindParameter("diag_Pgain", m_diag_Pgain, "1.0");
  bindParameter("diag_Dgain", m_diag_Dgain, "1.0");
  
  m_configsets.activateConfigurationSet("default");
  
  return RTC::RTC_OK;
}


RTC::ReturnCode_t PositionController::onFinalize()
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t PositionController::onActivated(RTC::UniqueId ec_id)
{
  // set PD gains
  if(m_diag_Pgain.size() != m_diag_Dgain.size() || 
     m_diag_Pgain.size() < 6 || m_diag_Dgain.size() < 6)
    return RTC::RTC_ERROR;
  
  DblMatrix Kp,Kd;
  unsigned int n = m_diag_Pgain.size();
  Kp.length(n);
  Kd.length(n);
  for (unsigned int i(0); i < n; i++) {
    Kp[i].length(n);
    Kd[i].length(n);
    for (unsigned int j(0); j < n; j++) {
      Kp[i][j] = 0.0;
      Kd[i][j] = 0.0;
    }
    Kp[i][i] = m_diag_Pgain[i];
    Kd[i][i] = m_diag_Dgain[i];
  }  
  
  if(CORBA::is_nil(m_PDControlSvc0.operator->())){
    std::cout<<"Service port not connected. Skipping controller setup."<<std::endl;
  }else{
    try {
      bool ret = m_PDControlSvc0->set_gains(Kp,Kd);
      if (!ret) {
	std::cout << "Error: set_gains() !" << std::endl;
	return RTC::RTC_ERROR;
      }
    }
    catch (...) {
      std::cout << "Is the ServicePort connected ?" << std::endl;
      return RTC::PRECONDITION_NOT_MET;
    }
  }
  
  try{
    m_period = 1.0/m_execContexts[ec_id]->get_rate();
  }catch(...){
    return RTC::RTC_ERROR;
  }
  m_tick = 0;
  m_path_init = false;
  
#ifdef GET_LOG
  m_log.open(LOGFILE);
  m_log<<"# Logfile of Position Controller"<<std::endl;
  m_log<<"# Time[s] TargetPos(vector) CurrentPos(vector)"<<std::endl;
#endif
  
  return RTC::RTC_OK;
}


RTC::ReturnCode_t PositionController::onDeactivated(RTC::UniqueId ec_id)
{
#ifdef GET_LOG
  m_log.close();
#endif
  m_path_init = false;
  return RTC::RTC_OK;
}


RTC::ReturnCode_t PositionController::onExecute(RTC::UniqueId ec_id)
{
  if(m_path_init){
    double t = m_period*(double)m_tick;
    unsigned int n = m_start.size();
    TimedDoubleSeq s;
    s.data.length(n);
    
    // Constant velocity path generation
    for(unsigned int i = 0; i < n; ++i){
      double r = t/m_Td[i];
      if(r < 1.0){
	s.data[i] = r*m_goal[i] + (1.0 - r)*m_start[i];
      }else{
	s.data[i] = m_goal[i];
      }
    }
    
    m_PositionOutput.write(s);
    
#ifdef GET_LOG
    m_log<<t; // time
    for(unsigned int i = 0; i < n; ++i) m_log<<" "<<s.data[i]; // target
    if(m_PositionInput.isNew()){
      s = m_PositionInput.read();
      n = s.data.length();
      for(unsigned int i = 0; i < n; ++i) m_log<<" "<<s.data[i]; // current
    }
#endif
    
    ++m_tick;
    
  }else{
    if(m_PositionInput.isNew()){
      TimedDoubleSeq s = m_PositionInput.read();
      CORBA::ULong n = s.data.length();
      m_start.resize(n);
      m_goal.resize(n);
      m_Td.resize(n);
      
      // Sample path
      for(CORBA::ULong i = 0; i < n; ++i){ 
	m_start[i] = s.data[i];
	m_goal[i] = m_start[i] + 0.5;
	m_Td[i] = 10.0;
      }
      m_tick = 0;
      m_path_init = true;
      
      std::cout<<"Path initialized"<<std::endl;
    }
  }
  
  return RTC::RTC_OK;
}



extern "C"
{
 
  void PositionControllerInit(RTC::Manager* manager)
  {
    RTC::Properties profile(positioncontroller_spec);
    manager->registerFactory(profile,
                             RTC::Create<PositionController>,
                             RTC::Delete<PositionController>);
  }
  
};



