// -*- C++ -*-
/*!
 * @file   FwdKinematicsPA10.cpp 
 * @brief  Forward Kinematics of PA10-7C arm
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/25
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "FwdKinematicsPA10.h"
#include "VecSeqConverter.h"

// Comment out next line in ordinary use
//#define DEBUG_MODE

// Macros
#define CERR(msg) std::cout<<"[FwdKinematics] ERROR: "<<msg<<std::endl;

#ifdef DEBUG_MODE
#define COUT(msg) std::cout<<"[FwdKinematics] "<<msg<<std::endl
#else
#define COUT(msg)
#endif

#define RETURN_ERROR(msg) do{ CERR(msg); return RTC::RTC_ERROR; }while(0)


namespace RTC {
  
  // Module specification
  static const char* fwdkinematicspa10_spec[] = {
    "implementation_id", "FwdKinematicsPA10",
    "type_name",         "FwdKinematicsPA10",
    "description",       "Forward Kinematics of PA10-7C arm",
    "version",           "1.0",
    "vendor",            "Masayuki Shimizu",
    "category",          "Kinematics",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.OrientationType", "Roll-Pitch-Yaw",
    ""
  };
  
  
  FwdKinematicsPA10::FwdKinematicsPA10(RTC::Manager* manager)
    : RTC::DataFlowComponentBase(manager),
      m_JointAngleInport("JointAngle", m_JointAngle),
      m_JointVelocityInport("JointVelocity", m_JointVelocity),
      m_TipPositionOutport("TipPosition", m_TipPosition),
      m_DupTipPositionOutport("DupTipPosition", m_DupTipPosition),
      m_TipVelocityOutport("TipVelocity", m_TipVelocity)
  {
    // Set InPort buffers
    registerInPort("Joint Angle", m_JointAngleInport);
    registerInPort("Joint Velocity", m_JointVelocityInport);
    
    // Set OutPort buffer
    registerOutPort("Tip Position", m_TipPositionOutport);
    registerOutPort("Duplicate Tip Position", m_DupTipPositionOutport);
    registerOutPort("Tip Velocity", m_TipVelocityOutport);
  }
  
  FwdKinematicsPA10::~FwdKinematicsPA10()
  {
  }
  
  
  RTC::ReturnCode_t FwdKinematicsPA10::onInitialize()
  {
    bindParameter("OrientationType", m_orientation_type, "Unknown");
    
    // apply "default" configset
    m_configsets.activateConfigurationSet("default");
    
    return RTC::RTC_OK;
  }

  RTC::ReturnCode_t FwdKinematicsPA10::onFinalize()
  {
    return RTC::RTC_OK;
  }
  
  RTC::ReturnCode_t FwdKinematicsPA10::onExecute(RTC::UniqueId ec_id)
  {
    if(m_JointAngleInport.isNew()){
      // joint angle -> tip position & orientation
      m_JointAngle = m_JointAngleInport.read();
      seq_to_vec(m_JointAngle, m_jangle);
      
      if(!m_kin.tip_pose(m_jangle, m_tip_pos, m_tip_rot))
	RETURN_ERROR("Forward kinematic computation failed.");
      
      if(!matrix_to_vector(m_tip_pos, m_tip_rot, m_tip_posture))
	RETURN_ERROR("Invalid orientation format, or singular orientation.");
      
      vec_to_seq(m_tip_posture, m_TipPosition);
      m_TipPositionOutport.write(m_TipPosition);
      vec_to_seq(m_tip_posture, m_DupTipPosition);
      m_DupTipPositionOutport.write(m_DupTipPosition);
      
      // joint velocity -> tip velocity
      if(m_JointVelocityInport.isNew()){
	m_JointVelocity = m_JointVelocityInport.read();
	seq_to_vec(m_JointVelocity, m_jspeed);
	
	if(!m_kin.tip_velocity(m_jangle, m_jspeed, m_tip_vel))
	  RETURN_ERROR("Jacobian computation failed.");
	
	vec_to_seq(m_tip_vel, m_TipVelocity);
	m_TipVelocityOutport.write(m_TipVelocity);
      }
    }
    
    return RTC::RTC_OK;
  }
};


extern "C"
{
 
  void FwdKinematicsPA10Init(RTC::Manager* manager)
  {
    RTC::Properties profile(RTC::fwdkinematicspa10_spec);
    manager->registerFactory(profile,
                             RTC::Create<RTC::FwdKinematicsPA10>,
                             RTC::Delete<RTC::FwdKinematicsPA10>);
  }
  
};



