// -*- C++ -*-
/*!
 * @file   FwdKinematicsPA10.h 
 * @brief  Forward Kinematics of PA10-7C arm
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/11/27
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef FWDKINEMATICSPA10_H
#define FWDKINEMATICSPA10_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

#include "fwd_kin_pa10.h"
#include "RotationConverter.h"


namespace RTC {
  
  class FwdKinematicsPA10  : public RTC::DataFlowComponentBase
  {
  public:
    FwdKinematicsPA10(RTC::Manager* manager);
    virtual ~FwdKinematicsPA10();
    
    virtual RTC::ReturnCode_t onInitialize();
    virtual RTC::ReturnCode_t onFinalize();
    
    virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);
    
  protected:
    TimedDoubleSeq m_JointAngle;
    InPort<TimedDoubleSeq> m_JointAngleInport;
    TimedDoubleSeq m_JointVelocity;
    InPort<TimedDoubleSeq> m_JointVelocityInport;
    
    TimedDoubleSeq m_TipPosition;
    OutPort<TimedDoubleSeq> m_TipPositionOutport;
    TimedDoubleSeq m_DupTipPosition;
    OutPort<TimedDoubleSeq> m_DupTipPositionOutport;
    TimedDoubleSeq m_TipVelocity;
    OutPort<TimedDoubleSeq> m_TipVelocityOutport;
    
  private:
    std::string m_orientation_type;
    
    fwd_kin_pa10 m_kin;
    std::vector<double> m_jangle;
    std::vector<double> m_jspeed;
    std::vector<double> m_tip_pos;
    std::vector<std::vector<double> > m_tip_rot;
    std::vector<double> m_tip_posture;
    std::vector<double> m_tip_vel;

    
    /**
     * @brief Convert position and orientation to 6-D vector.
     * @param[in] x Position vector (3-D).
     * @param[in] R Rotation matrix.
     * @param[out] pose 6-D vector representing position & orientation.
     */
    inline bool matrix_to_vector(const std::vector<double>& x, 
				 const std::vector<std::vector<double> >& R, 
				 std::vector<double>& pose)
    {
      if(x.size() < 3) return false;
      
      bool ret(false);
      std::vector<double> angle;
      
      // Convert rotation matrix to 3 angle parameters
      if(m_orientation_type == "Roll-Pitch-Yaw" ||
	 m_orientation_type == "roll-pitch-yaw" ||
	 m_orientation_type == "RPY" ||
	 m_orientation_type == "rpy"){
	ret = RotConv::matrix_to_rpy(R, angle);
      }
      
      if(!ret) return false;
      pose.resize(6);
      for(int i = 0; i < 3; ++i){
	pose[i] = x[i];
	pose[i+3] = angle[i];
      }
      
      return true;
    }
    
  };
  
};


extern "C"
{
  void FwdKinematicsPA10Init(RTC::Manager* manager);
};

#endif // FWDKINEMATICSPA10_H

