/**
 * @file   RotationConverter.h
 * @brief  Convert between rotation matrix and angle parameters.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/13
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 * 
 * @todo   Implement converter of euler angles. 
 */

#ifndef ROTATION_CONVERTER_H
#define ROTATION_CONVERTER_H

#include <vector>
#include <cmath>

#ifndef M_PI
#define M_PI (3.14159265358979323846)
#endif

#undef TOLERANCE
#define TOLERANCE (1.0e-16)

namespace RotConv {
  
  /**
   * @brief Convert roll-pitch-yaw angles to rotation matrix.
   * @param[in] rpy Roll-Pich-Yaw angles [rad].
   * @param[out] matrix Rotation matrix.
   */
  inline bool rpy_to_matrix(const std::vector<double>& rpy, 
			    std::vector<std::vector<double> >& matrix)
  {
    if(rpy.size() < 3) return false;
    
    double cr(cos(rpy[0])), sr(sin(rpy[0]));
    double cp(cos(rpy[1])), sp(sin(rpy[1]));
    double cy(cos(rpy[2])), sy(sin(rpy[2]));
    
    matrix.resize(3);
    for(int i = 0; i < 3; ++i) matrix[i].resize(3);
    
    matrix[0][0] = cy*cp;
    matrix[1][0] = sy*cp;
    matrix[2][0] =   -sp;
    
    matrix[0][1] = cy*sp*sr - sy*cr;
    matrix[1][1] = sy*sp*sr + cy*cr;
    matrix[2][1] =    cp*sr;
    
    matrix[0][2] = cy*sp*cr + sy*sr;
    matrix[1][2] = sy*sp*cr - cy*sr;
    matrix[2][2] =    cp*cr;
    
    return true;
  }
  
  /**
   * @brief Convert rotation matrix to roll-pitch-yaw angles.
   * @param[in] matrix Rotation matrix.
   * @param[out] rpy Roll-Pitch-Yaw angles [rad].
   * @param[in] cos_pitch_is_plus There are two possible RPY angles 
   *            for single rotation matrix. 
   *            One solution corresponds to cos(pitch) is plus, 
   *            the other is for negative value of cos(pitch). 
   *            If you want cos(pitch) to be positive, pass true. 
   *            Or, if you want cos(pitch) to be negative, pass false. 
   *            Default is true.
   */
  inline bool matrix_to_rpy(const std::vector<std::vector<double> >& matrix, 
			    std::vector<double>& rpy, 
			    bool cos_pitch_is_plus = true)
  {
    if(matrix.size() != 3) return false;
    for(int i = 0; i < 3; ++i) 
      if(matrix[i].size() != 3) return false;
    
    // check if singular
    if(fabs(matrix[0][0]) < TOLERANCE && fabs(matrix[1][0]) < TOLERANCE)
      return false;
    if(fabs(matrix[2][1]) < TOLERANCE && fabs(matrix[2][2]) < TOLERANCE)
      return false;
    
    rpy.resize(3);
    if(cos_pitch_is_plus){
      rpy[0] = atan2(matrix[2][1], matrix[2][2]);
      rpy[1] = -asin(matrix[2][0]);
      rpy[2] = atan2(matrix[1][0], matrix[0][0]);
    }else{
      rpy[0] = atan2(-matrix[2][1], -matrix[2][2]);
      rpy[1] = -asin(matrix[2][0]);
      rpy[2] = atan2(-matrix[1][0], -matrix[0][0]);
      rpy[1] = (rpy[1] < 0.0? -M_PI - rpy[1] : M_PI - rpy[1]);
    }
    
    return true;
  }
  
  
  /**
   * @brief Convert zyx euler angles to rotation matrix.
   * @param[in] zyx ZYX-euler angles [rad].
   * @param[out] matrix Rotation matrix.
   */
  inline bool zyx_to_matrix(const std::vector<double>& zyx, 
			    std::vector<std::vector<double> >& matrix)
  {
    if(zyx.size() < 3) return false;
    // Not implemented
    return false;
  }
  
  /**
   * @brief Convert rotation matrix to zyx euler angles.
   * @param[in] matrix Rotation matrix.
   * @param[out] zyx ZYX-euler angles [rad].
   */
  inline bool matrix_to_zyx(const std::vector<std::vector<double> >& matrix, 
			    std::vector<double>& zyx)
  {
    // Not implemented
    return false;
  }
  
};

#endif /* ROTATION_CONVERTER_H */
