/**
 * @file   fwd_kin_pa10.cpp
 * @brief  Forward Kinematics of PA10-7C arm.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/25
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "fwd_kin_pa10.h"
#include <cmath>
#include "matrix_util_inl.h"


/*=====================================
 *  Inline functions
 *=====================================*/
static inline void set_identity(std::vector<std::vector<double> >& x, 
				unsigned int size = 3)
{
  x.resize(size);
  for(unsigned int i = 0; i < size; ++i){
    x[i].resize(size);
    for(unsigned int j = 0; j < size; ++j) x[i][j] = 0.0;
    x[i][i] = 1.0;
  }
}


/*===================================================
 *   Ctor & Dtor
 *===================================================*/
fwd_kin_pa10::fwd_kin_pa10(const double base_to_shoulder, 
			   const double shoulder_to_elbow, 
			   const double elbow_to_wrist, 
			   const double wrist_to_tip)
  : lbs(base_to_shoulder), lse(shoulder_to_elbow),
    lew(elbow_to_wrist), lwt(wrist_to_tip)
{
  if(lbs < 0.0) throw int(-1);
  if(lse < 0.0) throw int(-2);
  if(lew < 0.0) throw int(-3);
  if(lwt < 0.0) throw int(-4);
}

fwd_kin_pa10::~fwd_kin_pa10()
{
}



/*=============================================================
 *   Forward kinematics
 *=============================================================*/
bool fwd_kin_pa10::tip_pose(const std::vector<double>& joint_angle, 
			    std::vector<double>& tip_position, 
			    std::vector<std::vector<double> >& tip_orientation)
{
  if(joint_angle.size() < 7) return false;
  
  std::vector<double> x(3, 0.0);
  std::vector<std::vector<double> > R, Rs, Re, Rw;
  R03(joint_angle, Rs);
  R34(joint_angle, Re);
  R47(joint_angle, Rw);
  
  for(unsigned int i = 0; i < 3; ++i) x[i] = lwt*Rw[i][2];
  x[2] += lew;
  x = mvprod(Re, x);
  x[1] -= lse;
  x = mvprod(Rs, x);
  x[2] += lbs;
  
  R = mmprod(Rs, Re);
  R = mmprod(R, Rw);
  
  tip_position = x;
  tip_orientation = R;
  
  return true;
}

bool fwd_kin_pa10::tip_velocity(const std::vector<double>& joint_angle, 
				const std::vector<double>& joint_velocity, 
				std::vector<double>& tip_velocity)
{
  std::vector<std::vector<double> > J;
  if(!jacobian(joint_angle, J)) return false;
  if(joint_velocity.size() != 7) return false;
  
  std::vector<double> V(6, 0.0);
  for(int i = 0; i < 6; ++i){
    for(int j = 0; j < 7; ++j)
      V[i] += J[i][j]*joint_velocity[j];
  }
  
  tip_velocity = V;
  
  return true;
}

bool fwd_kin_pa10::jacobian(const std::vector<double>& joint_angle, 
			    std::vector<std::vector<double> >& J)
{
  if(joint_angle.size() != 7) return false;
  
  double c1 = cos(joint_angle[0]);
  double s1 = sin(joint_angle[0]);
  double c2 = cos(joint_angle[1]);
  double s2 = sin(joint_angle[1]);
  double c3 = cos(joint_angle[2]);
  double s3 = sin(joint_angle[2]);
  double c4 = cos(joint_angle[3]);
  double s4 = sin(joint_angle[3]);
  double c5 = cos(joint_angle[4]);
  double s5 = sin(joint_angle[4]);
  double c6 = cos(joint_angle[5]);
  double s6 = sin(joint_angle[5]);
  
  std::vector<double> eXet(3, 0.0), sXst(3, 0.0), tmp(3, 0.0);
  std::vector<std::vector<double> > oRs, oRe, sRe, eRw;
  R03(joint_angle, oRs);
  R34(joint_angle, sRe); 
  oRe = mmprod(oRs, sRe);
  R47(joint_angle, eRw);
  for(int i = 0; i < 3; ++i) eXet[i] = lwt*eRw[i][2];
  eXet[2] += lew;
  sXst = mvprod(sRe, eXet);
  sXst[1] -= lse;
  
  J.resize(6);
  for(int i = 0; i < 6; ++i) J[i].resize(7);
  
  // linear velocity
  J[0][6] = 0.0;
  J[1][6] = 0.0;
  J[2][6] = 0.0;
  
  tmp[0] = lwt*c5*c6; 
  tmp[1] = lwt*s5*c6; 
  tmp[2] =   -lwt*s6;
  tmp = mvprod(oRe, tmp);
  J[0][5] = tmp[0];
  J[1][5] = tmp[1];
  J[2][5] = tmp[2];
  
  tmp[0] = -lwt*s5*s6; 
  tmp[1] =  lwt*c5*s6; 
  tmp[2] = 0.0;
  tmp = mvprod(oRe, tmp);
  J[0][4] = tmp[0];
  J[1][4] = tmp[1];
  J[2][4] = tmp[2];
  
  tmp[0] = -s4*eXet[0] + c4*eXet[2];
  tmp[1] =  c4*eXet[0] + s4*eXet[2];
  tmp[2] = 0.0;
  tmp = mvprod(oRs, tmp);
  J[0][3] = tmp[0];
  J[1][3] = tmp[1];
  J[2][3] = tmp[2];
  
  J[0][2] = (-c1*c2*s3 - s1*c3)*sXst[0] + (-c1*c2*c3 + s1*s3)*sXst[2];
  J[1][2] = (-s1*c2*s3 + c1*c3)*sXst[0] + (-s1*c2*c3 - c1*s3)*sXst[2];
  J[2][2] =               s2*s3*sXst[0] +               s2*c3*sXst[2];
  
  tmp[0] = c3*sXst[0] - s3*sXst[2];
  tmp[2] = -sXst[1];
  J[0][1] = -c1*s2*tmp[0] + c1*c2*tmp[2];
  J[1][1] = -s1*s2*tmp[0] + s1*c2*tmp[2];
  J[2][1] =    -c2*tmp[0] -    s2*tmp[2];
  
  tmp[0] = c2*c3*sXst[0] - s2*sXst[1] - c2*s3*sXst[2];
  tmp[2] =    s3*sXst[0]              +    c3*sXst[2];
  J[0][0] = -s1*tmp[0] - c1*tmp[2];
  J[1][0] =  c1*tmp[0] - s1*tmp[2];
  J[2][0] = 0.0;

  // angular velocity
  J[3][0] = 0.0;
  J[4][0] = 0.0;
  J[5][0] = 1.0;
  
  J[3][1] = -s1;
  J[4][1] =  c1;
  J[5][1] = 0.0;
  
  J[3][2] = c1*s2;
  J[4][2] = s1*s2;
  J[5][2] =    c2;
  
  J[3][3] = oRs[0][2];
  J[4][3] = oRs[1][2];
  J[5][3] = oRs[2][2];
  
  J[3][4] = oRe[0][2];
  J[4][4] = oRe[1][2];
  J[5][4] = oRe[2][2];
  
  tmp[0] = -s5; 
  tmp[1] =  c5; 
  tmp[2] = 0.0;
  tmp = mvprod(oRe, tmp);
  J[3][5] = tmp[0];
  J[4][5] = tmp[1];
  J[5][5] = tmp[2];
  
  tmp[0] = eRw[0][2]; 
  tmp[1] = eRw[1][2]; 
  tmp[2] = eRw[2][2];
  tmp = mvprod(oRe, tmp);
  J[3][6] = tmp[0];
  J[4][6] = tmp[1];
  J[5][6] = tmp[2];
  
  return true;
}



bool fwd_kin_pa10::R03(const std::vector<double>& jang, 
		       std::vector<std::vector<double> >& rot)
{
  if(jang.size() < 3) return false;
  set_identity(rot);
  
  double c1 = cos(jang[0]);
  double s1 = sin(jang[0]);
  double c2 = cos(jang[1]);
  double s2 = sin(jang[1]);
  double c3 = cos(jang[2]);
  double s3 = sin(jang[2]);

  // column 1
  rot[0][0] = c1*c2*c3 - s1*s3;
  rot[1][0] = s1*c2*c3 + c1*s3;
  rot[2][0] =   -s2*c3;

  // column 2
  rot[0][1] = -c1*s2;
  rot[1][1] = -s1*s2;
  rot[2][1] =    -c2;

  // column 3
  rot[0][2] = -c1*c2*s3 - s1*c3;
  rot[1][2] = -s1*c2*s3 + c1*c3;
  rot[2][2] =     s2*s3;
  
  return true;
}

bool fwd_kin_pa10::R34(const std::vector<double>& jang, 
		       std::vector<std::vector<double> >& rot)
{
  if(jang.size() < 4) return false;
  set_identity(rot);

  double c4 = cos(jang[3]);
  double s4 = sin(jang[3]);

   // column 1
  rot[0][0] =  c4;
  rot[1][0] =  s4;
  rot[2][0] = 0.0;
  
  // column 2
  rot[0][1] = 0.0;
  rot[1][1] = 0.0;
  rot[2][1] = 1.0;
  
  // column 3
  rot[0][2] =  s4;
  rot[1][2] = -c4;
  rot[2][2] = 0.0;

  return true;
}

bool fwd_kin_pa10::R47(const std::vector<double>& jang, 
		       std::vector<std::vector<double> >& rot)
{
  if(jang.size() < 7) return false;
  set_identity(rot);
  
  double c5 = cos(jang[4]);
  double s5 = sin(jang[4]);
  double c6 = cos(jang[5]);
  double s6 = sin(jang[5]);
  double c7 = cos(jang[6]);
  double s7 = sin(jang[6]);
  
  // column 1
  rot[0][0] =  c5*c6*c7 - s5*s7;
  rot[1][0] =  s5*c6*c7 + c5*s7;
  rot[2][0] =    -s6*c7;
  
  // column 2
  rot[0][1] = -c5*c6*s7 - s5*c7;
  rot[1][1] = -s5*c6*s7 + c5*c7;
  rot[2][1] =     s6*s7;
  
  // column 3
  rot[0][2] = c5*s6;
  rot[1][2] = s5*s6;
  rot[2][2] =    c6;
  
  return true;
}



/*================================================
 *   Configuration
 *================================================*/
bool fwd_kin_pa10::set_link_length(const double base_to_shoulder, 
				   const double shoulder_to_elbow, 
				   const double elbow_to_wrist, 
				   const double wrist_to_tip)
{
  if(base_to_shoulder < 0.0) return false;
  if(shoulder_to_elbow < 0.0) return false;
  if(elbow_to_wrist < 0.0) return false;
  if(wrist_to_tip < 0.0) return false;

  lbs = base_to_shoulder;
  lse = shoulder_to_elbow;
  lew = elbow_to_wrist;
  lwt = wrist_to_tip;

  return true;
}

void fwd_kin_pa10::get_link_length(double& base_to_shoulder, 
				   double& shoulder_to_elbow, 
				   double& elbow_to_wrist, 
				   double& wrist_to_tip)
{
  base_to_shoulder = lbs;
  shoulder_to_elbow = lse;
  elbow_to_wrist = lew;
  wrist_to_tip = lwt;
}
