/**
 * @file   fwd_kin_pa10.h
 * @brief  Forward kinematics of PA10-7C arm.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/27
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef FORWARD_KINEMATICS_PA10_H
#define FORWARD_KINEMATICS_PA10_H

#include <vector>


class fwd_kin_pa10{
 public:
  /**
   * @brief Constructor.
   * @param[in] base_to_shoulder Length between base & shoulder (m).
   * @param[in] shoulder_to_elbow Length between shoulder & elbow (m).
   * @param[in] elbow_to_wrist Length between elbow & wrist (m).
   * @param[in] wrist_to_tip Length between wrist & tip (m).
   * @note Default link lengths are PA10-7C's ones.
   */
  fwd_kin_pa10(const double base_to_shoulder = 0.317, 
	       const double shoulder_to_elbow = 0.450, 
	       const double elbow_to_wrist = 0.480, 
	       const double wrist_to_tip = 0.070);
  
  /**
   * @brief Destructor.
   */
  virtual ~fwd_kin_pa10();
  
  
  /**
   * @brief Compute tip position and orientation.
   * @param[in] joint_angle Joint angles [rad].
   * @param[out] tip_position Tip position vector [m].
   * @param[out] tip_orientation Rotation matrix of tip coordinate frame 
   *             relative to base coordinate frame.
   * @return - True on success.
   *         - False, if invalid values input.
   * @note The dimension of joint angle vector must be 7. 
   */
  virtual bool tip_pose(const std::vector<double>& joint_angle, 
			std::vector<double>& tip_position, 
			std::vector<std::vector<double> >& tip_orientation);
  
  /**
   * @brief Compute tip velocity.
   * @param[in] joint_angle Current joint angles [rad].
   * @param[in] joint_velocity Joint velocities [rad/s].
   * @param[out] tip_velocity Tip velocity (6 dimensional composite vector). 
   *             Upper 3-D vector is linear velocity [m/s], while
   *             lower 3-D vector is angular velocity [rad/s].
   * @return - True on success.
   *         - False, if invalid values input.
   * @note The dimensions of joint angle and velocity vectors must be 7.
   */
  virtual bool tip_velocity(const std::vector<double>& joint_angle, 
			    const std::vector<double>& joint_velocity, 
			    std::vector<double>& tip_velocity);
  
  /**
   * @brief Compute Jacobian matrix.
   * @param[in] joint_angle Current joint angles [rad].
   * @param[out] Jmatrix Jacobian matrix (6x7 matrix).
   *             Upper part corresponds to linear velocity, while
   *             lower part corresponds to angular velocity.
   * @return - True on success.
   *         - False, if invalid values input.
   * @note The dimension of joint angle vector must be 7. 
   */
  virtual bool jacobian(const std::vector<double>& joint_angle, 
			std::vector<std::vector<double> >& Jmatrix);
  
  
  /**
   * @brief Set link lengths.
   * @param[in] base_to_shoulder Length between base & shoulder (m).
   * @param[in] shoulder_to_elbow Length between shoulder & elbow (m).
   * @param[in] elbow_to_wrist Length between elbow & wrist (m).
   * @param[in] wrist_to_tip Length between wrist & tip (m).
   */
  virtual bool set_link_length(const double base_to_shoulder, 
			       const double shoulder_to_elbow, 
			       const double elbow_to_wrist, 
			       const double wrist_to_tip);

  /**
   * @brief Get link lengths.
   * @param[out] base_to_shoulder Length between base & shoulder (m).
   * @param[out] shoulder_to_elbow Length between shoulder & elbow (m).
   * @param[out] elbow_to_wrist Length between elbow & wrist (m).
   * @param[out] wrist_to_tip Length between wrist & tip (m).
   */
  virtual void get_link_length(double& base_to_shoulder, 
			       double& shoulder_to_elbow, 
			       double& elbow_to_wrist, 
			       double& wrist_to_tip);
  
 private:
  /**
   * @brief Rotation matrix of shoulder portion.
   */
  bool R03(const std::vector<double>& jang, 
	   std::vector<std::vector<double> >& rot);
  
  /**
   * @brief Rotation matrix of elbow portion.
   */
  bool R34(const std::vector<double>& jang, 
	   std::vector<std::vector<double> >& rot);
  
  /**
   * @brief Rotation matrix of wrist portion.
   */
  bool R47(const std::vector<double>& jang, 
	   std::vector<std::vector<double> >& rot);
  
  double lbs;
  double lse;
  double lew;
  double lwt;
};

#endif
