/**
 * @file    matrix_util_inl.h
 * @brief   Inline utilities for vector & matrix
 * @author  Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date    2008/10/26
 * @version 0.4
 * @note    Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef MATRIX_UTILITIES_INLINE_H
#define MATRIX_UTILITIES_INLINE_H

#include <cmath>


/**
 * @brief Check if matrix is square.
 */
inline bool is_square(const std::vector<std::vector<double> >& x, 
		      unsigned int size = 0)
{
  unsigned int n; 
  if(size != 0){
    n = size;
    if(x.size() != n) return false;
  }else{
    n = x.size();
    if(!n) return false;
  }
  for(unsigned int i = 0; i < n; ++i){
    if(x[i].size() != n) 
      return false;
  }
  return true;
}


/**
 * @brief Product of matrix and vector
 */
inline std::vector<double> mvprod(const std::vector<std::vector<double> >& m,
				  const std::vector<double>& v)
{
  std::vector<double> ret;
  
  unsigned int a = m.size();
  if(!a) return ret;
  unsigned int b = v.size();
  if(!b) return ret;
  
  unsigned int i,j;
  ret.resize(a);
  for(i = 0; i < a; ++i){
    if(m[i].size() != b){ ret.clear(); return ret; }
    ret[i] = 0.0;
    for(j = 0; j < b; ++j)
      ret[i] += m[i][j]*v[j];
  }

  return ret;
}


/**
 * @brief Product of two matrices
 */
inline std::vector<std::vector<double> > 
mmprod(const std::vector<std::vector<double> >& x, 
       const std::vector<std::vector<double> >& y)
{
  std::vector<std::vector<double> > ret;
  
  unsigned int a = x.size();
  if(!a) return ret;
  unsigned int b = y.size();
  if(!b) return ret;
  unsigned int c = y[0].size();
  if(!c) return ret;
  
  ret.resize(a);
  unsigned int i,j,k;
  for(i = 0; i < a; ++i){
    if(x[i].size() != b){ ret.clear(); return ret; }
    ret[i].resize(c);
    for(j = 0; j < c; ++j){
      ret[i][j] = 0.0;
      for(k = 0; k < b; ++k){ 
	if(y[k].size() != c){ ret.clear(); return ret; }
	ret[i][j] += x[i][k]*y[k][j];
      }
    }
  }
  
  return ret;
}


/**
 * @brief Matrix subtraction
 */
inline std::vector<std::vector<double> >
msub(const std::vector<std::vector<double> >& x, 
     const std::vector<std::vector<double> >& y)
{
  std::vector<std::vector<double> > ret;
  unsigned int m = x.size();
  if(m == 0 || y.size() != m) return ret;
  unsigned int n = x[0].size();
  unsigned int i,j;
  for(i = 0; i < m; ++i){
    if(x[i].size() != n || y[i].size() != n) return ret;
  }
  
  ret.resize(m);
  for(i = 0; i < m; ++i){
    ret[i].resize(n);
    for(j = 0; j < n; ++j) ret[i][j] = x[i][j] - y[i][j];
  }

  return ret;
}


/**
 * @brief Matrix addition
 */
inline std::vector<std::vector<double> >
madd(const std::vector<std::vector<double> >& x, 
     const std::vector<std::vector<double> >& y)
{
  std::vector<std::vector<double> > ret;
  unsigned int m = x.size();
  if(m == 0 || y.size() != m) return ret;
  unsigned int n = x[0].size();
  unsigned int i,j;
  for(i = 0; i < m; ++i){
    if(x[i].size() != n || y[i].size() != n) return ret;
  }
  
  ret.resize(m);
  for(i = 0; i < m; ++i){
    ret[i].resize(n);
    for(j = 0; j < n; ++j) ret[i][j] = x[i][j] + y[i][j];
  }
  
  return ret;
}


/**
 * @brief Matrix scaling
 */
inline std::vector<std::vector<double> >
mscale(const std::vector<std::vector<double> >& x, double coefficient)
{
  std::vector<std::vector<double> > ret;
  unsigned int m = x.size();
  unsigned int n = x[0].size();
  unsigned int i,j;
  for(i = 0; i < m; ++i) if(x[i].size() != n) return ret;
  
  ret.resize(m);
  for(i = 0; i < m; ++i){
    ret[i].resize(n);
    for(j = 0; j < n; ++j) ret[i][j] = coefficient*x[i][j];
  }

  return ret;
}


/**
 * @brief Inner product of vector
 */
inline double inner_prod(const std::vector<double>& a, 
			 const std::vector<double>& b)
{
  unsigned int n = a.size();
  double ret = 0.0;
  if(b.size() != n) return ret;
  for(unsigned int i = 0; i < n; ++i)
    ret += a[i]*b[i];
  return ret;
}

/**
 * @brief Square of norm of vector
 */
inline double norm2(const std::vector<double>& v)
{
  unsigned int n = v.size();
  double ret = 0.0;
  for(unsigned int i = 0; i < n; ++i)
    ret += v[i]*v[i];
  return ret;
}

/**
 * @brief Norm of vector
 */
inline double norm(const std::vector<double>& v)
{
  return sqrt(norm2(v));
}


/**
 * @brief Outer product of 3-D vectors
 */
inline std::vector<double> outer_prod(const std::vector<double>& a, 
				      const std::vector<double>& b)
{
  std::vector<double> ret(3, 0.0);
  if(a.size() < 3 || b.size() < 3){ ret.clear(); return ret; }
  
  ret[0] =            - a[2]*b[1] + a[1]*b[2];
  ret[1] =  a[2]*b[0]             - a[0]*b[2];
  ret[2] = -a[1]*b[0] + a[0]*b[1];
  
  return ret;
}


/**
 * @brief Inverse of 3x3 matrix
 */
inline std::vector<std::vector<double> > 
inverse33(const std::vector<std::vector<double> >& A, 
	  double tolerance = 1.0e-16)
{
  unsigned int i;
  std::vector<std::vector<double> > Ainv;
  if(A.size() < 3) return Ainv;
  for(i = 0; i < 3; ++i) if(A[i].size() < 3) return Ainv;
  
  // check if singular
  double d,c1[3];
  c1[0] = A[1][1]*A[2][2] - A[1][2]*A[2][1];
  c1[1] = A[1][2]*A[2][0] - A[1][0]*A[2][2];
  c1[2] = A[1][0]*A[2][1] - A[1][1]*A[2][0];
  d = A[0][0]*c1[0] + A[0][1]*c1[1] + A[0][2]*c1[2];
  if(fabs(d) < tolerance) return Ainv;
  
  double c2[3],c3[3];
  c2[0] = (A[0][2]*A[2][1] - A[0][1]*A[2][2]);
  c2[1] = (A[0][0]*A[2][2] - A[0][2]*A[2][0]);
  c2[2] = (A[0][1]*A[2][0] - A[0][0]*A[2][1]);
  c3[0] = (A[0][1]*A[1][2] - A[0][2]*A[1][1]);
  c3[1] = (A[0][2]*A[1][0] - A[0][0]*A[1][2]);
  c3[2] = (A[0][0]*A[1][1] - A[0][1]*A[1][0]);
  
  Ainv.resize(3);
  for(i = 0; i < 3; ++i){ 
    Ainv[i].resize(3);
    Ainv[i][0] = c1[i]/d;
    Ainv[i][1] = c2[i]/d;
    Ainv[i][2] = c3[i]/d;
  }
  
  return Ainv;
}


/**
 * @brief Transpose of matrix
 */
inline std::vector<std::vector<double> > 
transpose(const std::vector<std::vector<double> >& A)
{
  std::vector<std::vector<double> > ret;
  unsigned int i, j, n = A.size();
  for(i = 0; i < n; ++i) if(A[i].size() != n) return ret;
  ret.resize(n);
  for(i = 0; i < n; ++i){
    ret[i].resize(n);
    for(j = 0; j < n; ++j) ret[i][j] = A[j][i];
  }
  return ret;
}





#endif
