// -*- C++ -*-
/*!
 * @file   InvKinematicsPA10.h 
 * @brief  Inverse Kinematics of PA10-7C arm
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/11/27
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "InvKinematicsPA10.h"
#include "VecSeqConverter.h"


// Comment out next line in ordinary use
//#define DEBUG_MODE

// Macros
#define CERR(msg) std::cout<<"[InvKinematics] ERROR: "<<msg<<std::endl;

#ifdef DEBUG_MODE
#define COUT(msg) std::cout<<"[InvKinematics] "<<msg<<std::endl
#else
#define COUT(msg)
#endif

#define RETURN_ERROR(msg) do{ CERR(msg); return RTC::RTC_ERROR; }while(0)


namespace RTC {
  
  // Module specification
  static const char* invkinematicspa10_spec[] = {
    "implementation_id", "InvKinematicsPA10",
    "type_name",         "InvKinematicsPA10",
    "description",       "Inverse Kinematics of PA10-7C arm",
    "version",           "1.0",
    "vendor",            "Masayuki Shimizu",
    "category",          "Kinematics",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.OrientationType", "Roll-Pitch-Yaw", 
    ""
  };
  
  
  InvKinematicsPA10::InvKinematicsPA10(RTC::Manager* manager)
    : RTC::DataFlowComponentBase(manager),
      m_TipPositionInport("TipPosition", m_TipPosition),
      m_TipVelocityInport("TipVelocity", m_TipVelocity),
      m_RefJointAngleInport("RefJointAngle", m_RefJointAngle),
      m_JointAngleOutport("JointAngle", m_JointAngle),
      m_JointVelocityOutport("JointVelocity", m_JointVelocity)
  {
    registerInPort("Tip Position", m_TipPositionInport);
    registerInPort("Tip Velocity", m_TipVelocityInport);
    registerInPort("Reference Joint Angle", m_RefJointAngleInport);
    
    registerOutPort("Joint Angle", m_JointAngleOutport);
    registerOutPort("Joint Velocity", m_JointVelocityOutport);
  }
  
  InvKinematicsPA10::~InvKinematicsPA10()
  {
  }
  
  
  RTC::ReturnCode_t InvKinematicsPA10::onInitialize()
  {
    bindParameter("OrientationType", m_orientation_type, "Unknown");
    
    // apply "default" configset
    m_configsets.activateConfigurationSet("default");
    
    return RTC::RTC_OK;
  }
  
  RTC::ReturnCode_t InvKinematicsPA10::onFinalize()
  {
    return RTC::RTC_OK;
  }
  
  
  RTC::ReturnCode_t InvKinematicsPA10::onActivated(RTC::UniqueId ec_id)
  {
    m_refang.clear();
    m_jangle.clear();
    m_jspeed.clear();
    
    return RTC::RTC_OK;
  }

  RTC::ReturnCode_t InvKinematicsPA10::onDeactivated(RTC::UniqueId ec_id)
  {
    return RTC::RTC_OK;
  }
  

  RTC::ReturnCode_t InvKinematicsPA10::onExecute(RTC::UniqueId ec_id)
  {
    bool ang_updated = false;
    bool vel_updated = false;
    bool ref_updated = false;
    
    // update tip pose
    if(m_TipPositionInport.isNew()){
      m_TipPosition = m_TipPositionInport.read();
      seq_to_vec(m_TipPosition, m_tpose);
      if(!vector_to_matrix(m_tpose, m_xt, m_Rt))
	RETURN_ERROR("Perhaps singular orientation");
      ang_updated = true;
    }
    
    // update tip velocity
    if(m_TipVelocityInport.isNew()){
      m_TipVelocity = m_TipVelocityInport.read();
      seq_to_vec(m_TipVelocity, m_tvel);
      vel_updated = true;
    }
    
    // update reference joint angle
    if(m_RefJointAngleInport.isNew()){
      m_RefJointAngle = m_RefJointAngleInport.read();
      seq_to_vec(m_RefJointAngle, m_refang);
      ref_updated = true;
    }
    
    // compute joint angle from tip pose
    if(ang_updated){
      if(m_refang.size() >= 7){
	if(!m_kin.joint_angle_opt(m_xt, m_Rt, m_refang, m_jangle))
	  RETURN_ERROR("Failed to compute joint angle.");
      }else{
	if(!m_kin.joint_angle(m_xt, m_Rt, m_jangle))
	  RETURN_ERROR("Failed to compute joint angle.");
      }
      
      vec_to_seq(m_jangle, m_JointAngle);
      m_JointAngleOutport.write(m_JointAngle);
    }
    
    // compute joint velocity from tip velocity
    if(vel_updated){
      if(ref_updated){
	if(!m_kin.joint_velocity(m_tvel, m_refang, m_jspeed))
	  RETURN_ERROR("Failed to compute joint velocity.");
      }else if(ang_updated){
	if(!m_kin.joint_velocity(m_tvel, m_jangle, m_jspeed))
	  RETURN_ERROR("Failed to compute joint velocity.");
      }else{
	COUT("Joint angle/Tip pose is required to compute joint velocity.");
	return RTC::RTC_OK;
      }

      vec_to_seq(m_jspeed, m_JointVelocity);
      m_JointVelocityOutport.write(m_JointVelocity);
    }
    
    return RTC::RTC_OK;
  }
  
  
};


extern "C"
{
  
  void InvKinematicsPA10Init(RTC::Manager* manager)
  {
    RTC::Properties profile(RTC::invkinematicspa10_spec);
    manager->registerFactory(profile,
                             RTC::Create<RTC::InvKinematicsPA10>,
                             RTC::Delete<RTC::InvKinematicsPA10>);
  }
  
};



