// -*- C++ -*-
/*!
 * @file   InvKinematicsPA10.h 
 * @brief  Inverse Kinematics of PA10-7C arm
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/11/27
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef INVKINEMATICSPA10_H
#define INVKINEMATICSPA10_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

#include "inv_kin_pa10.h"
#include "RotationConverter.h"


namespace RTC {
  
  class InvKinematicsPA10  : public RTC::DataFlowComponentBase 
  {
  public:
    InvKinematicsPA10(RTC::Manager* manager);
    virtual ~InvKinematicsPA10();
    
    virtual RTC::ReturnCode_t onInitialize();
    virtual RTC::ReturnCode_t onFinalize();
    
    virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);
    virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);
    
    virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);
    
  protected:
    TimedDoubleSeq m_TipPosition;
    InPort<TimedDoubleSeq> m_TipPositionInport;
    TimedDoubleSeq m_TipVelocity;
    InPort<TimedDoubleSeq> m_TipVelocityInport;
    TimedDoubleSeq m_RefJointAngle;
    InPort<TimedDoubleSeq> m_RefJointAngleInport;
    
    TimedDoubleSeq m_JointAngle;
    OutPort<TimedDoubleSeq> m_JointAngleOutport;
    TimedDoubleSeq m_JointVelocity;
    OutPort<TimedDoubleSeq> m_JointVelocityOutport;
    
  private:
    std::string m_orientation_type;
    
    inv_kin_pa10 m_kin;
    std::vector<double> m_tpose;
    std::vector<double> m_tvel;
    std::vector<double> m_refang;

    std::vector<double> m_xt;
    std::vector<std::vector<double> > m_Rt;
    
    std::vector<double> m_jangle;
    std::vector<double> m_jspeed;

    
    /**
     * @brief Convert 6-D pose vector to position & orientation.
     * @param[in] pose 6-D pose vector.
     * @param[out] x 3-D position vector.
     * @param[out] R Rotation matrix. 
     */
    inline bool vector_to_matrix(const std::vector<double>& pose, 
				 std::vector<double>& x, 
				 std::vector<std::vector<double> >& R)
    {
      if(pose.size() < 6) return false;
      
      bool ret(false);
      std::vector<double> angle(3, 0.0);
      x.resize(3);
      for(int i = 0; i < 3; ++i){
	x[i] = pose[i];
	angle[i] = pose[i+3];
      }
      
      // Convert 3 angle parameters to rotation matrix
      if(m_orientation_type == "Roll-Pitch-Yaw" ||
	 m_orientation_type == "roll-pitch-yaw" ||
	 m_orientation_type == "RPY" ||
	 m_orientation_type == "rpy"){
	ret = RotConv::rpy_to_matrix(angle, R);
      }
      
      return ret;
    }
    
  };
};


extern "C"
{
  void InvKinematicsPA10Init(RTC::Manager* manager);
};

#endif // INVKINEMATICSPA10_H

