/**
 * @file   inv_kin_pa10.cpp
 * @brief  Inverse kinematics of PA10-7C arm.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/11/26
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "inv_kin_pa10.h"
#include "matrix_util_inl.h"
#include <cmath>

#include <iostream>

#ifndef TOLERANCE
#define TOLERANCE (1.0e-16)
#endif


/*=========================================
 *   Ctor & Dtor
 *=========================================*/
inv_kin_pa10::inv_kin_pa10(const double base_to_shoulder, 
			   const double shoulder_to_elbow, 
			   const double elbow_to_wrist, 
			   const double wrist_to_tip)
  : lbs(base_to_shoulder), lse(shoulder_to_elbow),
    lew(elbow_to_wrist), lwt(wrist_to_tip), 
    //qb(7, 0.0), 
    fwdkin(base_to_shoulder, shoulder_to_elbow, elbow_to_wrist, wrist_to_tip)
{
  if(lbs < 0.0) throw int(-1);
  if(lse < 0.0) throw int(-2);
  if(lew < 0.0) throw int(-3);
  if(lwt < 0.0) throw int(-4);
}

inv_kin_pa10::~inv_kin_pa10()
{
}



/*================================================
 *   Inverse kinematics
 *================================================*/
bool inv_kin_pa10::joint_angle(const std::vector<double>& tpos, 
			       const std::vector<std::vector<double> >& trot,
			       std::vector<double>& jang, 
			       unsigned int id)
{
  // sanity check
  if(tpos.size() != 3) return false;
  if(trot.size() != 3) return false;
  for(int i = 0; i < 3; ++i) if(trot[i].size() != 3) return false;
  if(id > 7) return false;
  
  bool ret(false);
  std::vector<double> q(7, 0.0);
  std::vector<double> xsw = stow(tpos, trot);
  
  ret = ((id & 0x04)? je(xsw, q, false) : je(xsw, q, true));
  if(!ret) return false; // elbow singularity
  
  ret = ((id & 0x02)? js(xsw, q, false) : js(xsw, q, true));
  if(!ret) return false; // shoulder singularity
  
  ret = ((id & 0x01)? jw(trot, q, false) : jw(trot, q, true));
  if(!ret) return false; // wrist singularity
  
  jang = q;
  
  return true;
}

bool inv_kin_pa10::joint_angle_opt(const std::vector<double>& tpos, 
				   const std::vector<std::vector<double> >& trot,
				   const std::vector<double>& rang, 
				   std::vector<double>& jang)
{
  // sanity check
  if(tpos.size() != 3) return false;
  if(trot.size() != 3) return false;
  for(int i = 0; i < 3; ++i) if(trot[i].size() != 3) return false;
  if(rang.size() < 7) return false;
  
  bool ret(false);
  std::vector<double> q(7, 0.0);
  std::vector<double> xsw = stow(tpos, trot);
  
  ret = je_r(xsw, rang, q);
  if(!ret) return false;
  
  ret = js_r(xsw, rang, q);
  if(!ret) return false;

  ret = jw_r(trot, rang, q);
  if(!ret) return false;

  jang = q;

  return true;
}

bool inv_kin_pa10::joint_velocity(const std::vector<double>& tvel, 
				  const std::vector<double>& jang, 
				  std::vector<double>& jvel)
{
  // sanity check
  if(tvel.size() != 6) return false;
  if(jang.size() != 7) return false;
  
  std::vector<std::vector<double> > J;
  if(!fwdkin.jacobian(jang, J)) return false;
  
  unsigned int i;
  std::vector<std::vector<double> > rJ(6);
  for(i = 0; i < 6; ++i){
    rJ[i].resize(6);
    rJ[i][0] = J[i][0];
    rJ[i][1] = J[i][1];
    rJ[i][2] = J[i][3];
    rJ[i][3] = J[i][4];
    rJ[i][4] = J[i][5];
    rJ[i][5] = J[i][6];
  }
  
  std::vector<std::vector<double> > rJinv;
  if(!invredJ(rJ, rJinv)) return false;
  std::vector<std::vector<double> > Jinv(7);
  for(i = 0; i < 7; ++i) Jinv[i].resize(6);
  for(i = 0; i < 6; ++i){
    Jinv[0][i] = rJinv[0][i];
    Jinv[1][i] = rJinv[1][i];
    Jinv[2][i] = 0.0;
    Jinv[3][i] = rJinv[2][i];
    Jinv[4][i] = rJinv[3][i];
    Jinv[5][i] = rJinv[4][i];
    Jinv[6][i] = rJinv[5][i];
  }
  jvel = mvprod(Jinv, tvel);
  
  return true;
}

std::vector<double> inv_kin_pa10::stow(const std::vector<double>& xt,
				       const std::vector<std::vector<double> >& Rt)
{
  std::vector<double> ret(3, 0.0);
  for(unsigned int i = 0; i < 3; ++i) ret[i] = xt[i] - lwt*Rt[i][2];
  ret[2] -= lbs;
  return ret;
}

bool inv_kin_pa10::je(const std::vector<double>& xsw, std::vector<double>& q, 
		      bool positive)
{
  double c = (norm2(xsw) - lse*lse - lew*lew)/(2.0*lse*lew);
  if(c <= -1.0 || c >= 1.0) return false;
  q[3] = (positive? acos(c) : -acos(c));
  return true;
}

bool inv_kin_pa10::js(const std::vector<double>& xsw, std::vector<double>& q, 
		      bool higher)
{
  std::vector<double> y(3, 0.0);
  y[0] = lew*sin(q[3]);
  y[1] = -lse - lew*cos(q[3]);
  
  q[2] = 0.0;
  
  {
    double r = y[0]*y[0] + y[1]*y[1] - xsw[2]*xsw[2];
    if(r < 0.0) return false;
    if(fabs(y[1] - xsw[2]) < TOLERANCE) return false;
    double tp = (y[0] + sqrt(r))/(y[1] - xsw[2]);
    double tm = (y[0] - sqrt(r))/(y[1] - xsw[2]);
    double cp = (-tp*y[0] - xsw[2])/(tp*y[0] + y[1]);
    double cm = (-tm*y[0] - xsw[2])/(tm*y[0] + y[1]);
    if(higher) q[1] = 2.0*(cp >= cm? atan(tp) : atan(tm));
    else q[1] = 2.0*(cp < cm? atan(tp) : atan(tm));
  }
  
  {
    double d = cos(q[1])*y[0] - sin(q[1])*y[1];
    if(fabs(d) < TOLERANCE) return false;
    double c = xsw[0]/d;
    double s = xsw[1]/d;
    q[0] = atan2(s,c);
  }

  return true;
}

bool inv_kin_pa10::jw(const std::vector<std::vector<double> >& Rt, 
		      std::vector<double>& q, bool nearer)
{
  unsigned int i;
  std::vector<std::vector<double> > ReT(3);
  for(i = 0; i < 3; ++i) ReT[i].resize(3);
  {
    double c1(cos(q[0])), s1(sin(q[0]));
    double c2(cos(q[1])), s2(sin(q[1]));
    double c4(cos(q[3])), s4(sin(q[3]));
    
    ReT[0][0] = c1*c2*c4 - c1*s2*s4;
    ReT[0][1] = s1*c2*c4 - s1*s2*s4;
    ReT[0][2] =   -s2*c4 -    c2*s4;

    ReT[1][0] = -s1;
    ReT[1][1] = c1;
    ReT[1][2] = 0.0;

    ReT[2][0] = c1*c2*s4 + c1*s2*c4;
    ReT[2][1] = s1*c2*s4 + s1*s2*c4;
    ReT[2][2] =   -s2*s4 +    c2*c4;
  }
  std::vector<std::vector<double> > Rw = mmprod(ReT, Rt);
  if(fabs(1.0 - Rw[2][2]) < TOLERANCE) return false;
  
  std::vector<double> qp(3, 0.0);
  qp[0] = atan2(Rw[1][2], Rw[0][2]);
  qp[1] = acos(Rw[2][2]);
  qp[2] = atan2(Rw[2][1], -Rw[2][0]);
  double cp(0.0); 
  for(i = 0; i < 3; ++i) 
    if(fabs(qp[i]) > cp) cp = fabs(qp[i]);
  
  std::vector<double> qm(3, 0.0);
  qm[0] = atan2(-Rw[1][2], -Rw[0][2]);
  qm[1] = -qp[1];
  qm[2] = atan2(-Rw[2][1], Rw[2][0]);
  double cm(0.0); 
  for(i = 0; i < 3; ++i) 
    if(fabs(qm[i]) > cm) cm = fabs(qm[i]);
  
  if(nearer){
    if(cp <= cm) for(i = 0; i < 3; ++i) q[i+4] = qp[i];
    else for(i = 0; i < 3; ++i) q[i+4] = qm[i];
  }else{
    if(cp > cm) for(i = 0; i < 3; ++i) q[i+4] = qp[i];
    else for(i = 0; i < 3; ++i) q[i+4] = qm[i];
  }
  
  return true;
}

bool inv_kin_pa10::je_r(const std::vector<double>& xsw, const std::vector<double>& qr,
			std::vector<double>& q)
{
  double c = (norm2(xsw) - lse*lse - lew*lew)/(2.0*lse*lew);
  if(c <= -1.0 || c >= 1.0) return false;
  q[3] = (qr[3] >= 0.0? acos(c) : -acos(c));
  return true;
}

bool inv_kin_pa10::js_r(const std::vector<double>& xsw, const std::vector<double>& qr, 
			std::vector<double>& q)
{
  std::vector<double> y(3, 0.0);
  y[0] = lew*sin(q[3]);
  y[1] = -lse - lew*cos(q[3]);
  
  q[2] = 0.0;
  
  double r = y[0]*y[0] + y[1]*y[1] - xsw[2]*xsw[2];
  if(r < 0.0) return false;
  if(fabs(y[1] - xsw[2]) < TOLERANCE) return false;
  double qp2 = 2.0*atan((y[0] + sqrt(r))/(y[1] - xsw[2]));
  double qm2 = 2.0*atan((y[0] - sqrt(r))/(y[1] - xsw[2]));
  
  double sp = cos(qp2)*y[0] - sin(qp2)*y[1];
  double sm = cos(qm2)*y[0] - sin(qm2)*y[1];
  double qp1 = (sp > 0.0? atan2(xsw[1], xsw[0]) : atan2(-xsw[1], -xsw[0]));
  double qm1 = (sm > 0.0? atan2(xsw[1], xsw[0]) : atan2(-xsw[1], -xsw[0]));
  
  double dp = fabs(qp2 - qr[1]) + fabs(qp1 - qr[0]); // total traveling distance
  double dm = fabs(qm2 - qr[1]) + fabs(qm1 - qr[0]);

  if(dp < dm){
    q[1] = qp2; q[0] = qp1;
  }else{
    q[1] = qm2; q[0] = qm1;
  }
  
  return true;
}

bool inv_kin_pa10::jw_r(const std::vector<std::vector<double> >& Rt, 
			const std::vector<double>& qr, 
			std::vector<double>& q)
{
  unsigned int i;
  std::vector<std::vector<double> > ReT(3);
  for(i = 0; i < 3; ++i) ReT[i].resize(3);
  {
    double c1(cos(q[0])), s1(sin(q[0]));
    double c2(cos(q[1])), s2(sin(q[1]));
    double c4(cos(q[3])), s4(sin(q[3]));
    
    ReT[0][0] = c1*c2*c4 - c1*s2*s4;
    ReT[0][1] = s1*c2*c4 - s1*s2*s4;
    ReT[0][2] =   -s2*c4 -    c2*s4;

    ReT[1][0] = -s1;
    ReT[1][1] = c1;
    ReT[1][2] = 0.0;

    ReT[2][0] = c1*c2*s4 + c1*s2*c4;
    ReT[2][1] = s1*c2*s4 + s1*s2*c4;
    ReT[2][2] =   -s2*s4 +    c2*c4;
  }
  std::vector<std::vector<double> > Rw = mmprod(ReT, Rt);
  if(fabs(1.0 - Rw[2][2]) < TOLERANCE) return false;
  
  std::vector<double> qp(3, 0.0);
  qp[0] = atan2(Rw[1][2], Rw[0][2]);
  qp[1] = acos(Rw[2][2]);
  qp[2] = atan2(Rw[2][1], -Rw[2][0]);
  double cp(0.0); 
  for(i = 0; i < 3; ++i) 
    if(fabs(qp[i] - qr[i+4]) > cp) cp = fabs(qp[i] - qr[i+4]);
  
  std::vector<double> qm(3, 0.0);
  qm[0] = atan2(-Rw[1][2], -Rw[0][2]);
  qm[1] = -qp[1];
  qm[2] = atan2(-Rw[2][1], Rw[2][0]);
  double cm(0.0); 
  for(i = 0; i < 3; ++i) 
    if(fabs(qm[i] - qr[i+4]) > cm) cm = fabs(qm[i] - qr[i+4]);
  
  if(cp < cm) for(i = 0; i < 3; ++i) q[i+4] = qp[i];
  else for(i = 0; i < 3; ++i) q[i+4] = qm[i];
  
  return true;
}

bool inv_kin_pa10::invredJ(const std::vector<std::vector<double> >& J, 
			   std::vector<std::vector<double> >& Jinv)
{
  unsigned int i, j;
  std::vector<std::vector<double> > A(3), B(3), C(3), D(3);
  for(i = 0; i < 3; ++i){
    A[i].resize(3);
    B[i].resize(3);
    C[i].resize(3);
    D[i].resize(3);
    for(j = 0; j < 3; ++j){
      A[i][j] = J[i][j];
      B[i][j] = J[i][j+3];
      C[i][j] = J[i+3][j];
      D[i][j] = J[i+3][j+3];
    }
  }

  std::vector<std::vector<double> > Dinv = inverse33(D);
  if(Dinv.size() != 3) return false; 
  std::vector<std::vector<double> > BDinv = mmprod(B, Dinv);
  std::vector<std::vector<double> > E1 = mmprod(BDinv, C);
  E1 = msub(A, E1);
  E1 = inverse33(E1);
  if(E1.size() != 3) return false;
  std::vector<std::vector<double> > E2 = mmprod(E1, BDinv);
  std::vector<std::vector<double> > DinvC = mmprod(Dinv, C);
  std::vector<std::vector<double> > E3 = mmprod(DinvC, E1);
  std::vector<std::vector<double> > E4 = mmprod(E3, BDinv);

  Jinv.resize(6);
  for(i = 0; i < 6; ++i) Jinv[i].resize(6);
  for(i = 0; i < 3; ++i)
    for(j = 0; j < 3; ++j){
      Jinv[i][j] = E1[i][j];
      Jinv[i][j+3] = -E2[i][j];
      Jinv[i+3][j] = -E3[i][j];
      Jinv[i+3][j+3] = Dinv[i][j] + E4[i][j];
    }
  
  return true;
}



/*================================================
 *   Configuration
 *================================================*/
bool inv_kin_pa10::set_link_length(const double base_to_shoulder, 
				   const double shoulder_to_elbow, 
				   const double elbow_to_wrist, 
				   const double wrist_to_tip)
{
  if(base_to_shoulder < 0.0) return false;
  if(shoulder_to_elbow < 0.0) return false;
  if(elbow_to_wrist < 0.0) return false;
  if(wrist_to_tip < 0.0) return false;
  
  lbs = base_to_shoulder;
  lse = shoulder_to_elbow;
  lew = elbow_to_wrist;
  lwt = wrist_to_tip;

  return true;
}

void inv_kin_pa10::get_link_length(double& base_to_shoulder, 
				   double& shoulder_to_elbow, 
				   double& elbow_to_wrist, 
				   double& wrist_to_tip)
{
  base_to_shoulder = lbs;
  shoulder_to_elbow = lse;
  elbow_to_wrist = lew;
  wrist_to_tip = lwt;
}
