/**
 * @file   inv_kin_pa10.h
 * @brief  Inverse kinematics of PA10-7C arm.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/11/26
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef INVERSE_KINEMATICS_PA10_H
#define INVERSE_KINEMATICS_PA10_H

#include <vector>
#include "fwd_kin_pa10.h"



class inv_kin_pa10{
 public:
  /**
   * @brief Constructor.
   * @param[in] base_to_shoulder Length between base & shoulder (m).
   * @param[in] shoulder_to_elbow Length between shoulder & elbow (m).
   * @param[in] elbow_to_wrist Length between elbow & wrist (m).
   * @param[in] wrist_to_tip Length between wrist & tip (m).
   * @note Default link lengths are PA10-7C's ones.
   */
  inv_kin_pa10(const double base_to_shoulder = 0.317, 
	       const double shoulder_to_elbow = 0.450, 
	       const double elbow_to_wrist = 0.480, 
	       const double wrist_to_tip = 0.070);
  
  /**
   * @brief Destructor.
   */
  virtual ~inv_kin_pa10();
  
  
  /**
   * @brief Compute joint angles for given position and orientation.
   * @param[in] tip_position Tip position vector [m].
   * @param[in] tip_orientation Rotation matrix of tip coordinate frame 
   *             relative to base coordinate frame.
   * @param[out] joint_angle Joint angles [rad].
   * @param[in] solution_id Joint angle solution id number (0-7). 
   *            There are eight inverse solutions for same tip pose. 
   *            Select suitable solution among the eight ones. 
   *            Default solution id number is 0, 
   *            which will be the best solution for most cases.
   * @return - True on success.
   *         - False, if invalid values input.
   *         - False, if singular configuration. 
   * @note  This computes inverse solutions for virtual non-redundant manipulator, 
   *        where joint #3 is fixed.
   */
  virtual bool joint_angle(const std::vector<double>& tip_position, 
			   const std::vector<std::vector<double> >& tip_orientation,
			   std::vector<double>& joint_angle, 
			   unsigned int solution_id = 0);
  
  /**
   * @brief Compute optimal joint angles for given position & orientation.
   */
  virtual bool joint_angle_opt(const std::vector<double>& tip_position, 
			       const std::vector<std::vector<double> >& tip_orientation,
			       const std::vector<double>& ref_angle, 
			       std::vector<double>& joint_angle);
  
  /**
   * @brief Compute joint velocities for given tip velocities.
   * @param[in] tip_velocity Tip velocity (6 dimensional composite vector). 
   *             Upper 3-D vector is linear velocity [m/s], while
   *             lower 3-D vector is angular velocity [rad/s].
   * @param[in] joint_angle Current joint angles [rad].
   * @param[out] joint_velocity Joint velocities [rad/s].
   * @return - True on success.
   *         - False, if invalid values input.
   *         - False, if singular configuration.
   * @note  Input variable joint_angle[2] must be zero. 
   */
  virtual bool joint_velocity(const std::vector<double>& tip_velocity, 
			      const std::vector<double>& joint_angle, 
			      std::vector<double>& joint_velocity);
  
  
  /**
   * @brief Set link lengths.
   * @param[in] base_to_shoulder Length between base & shoulder (m).
   * @param[in] shoulder_to_elbow Length between shoulder & elbow (m).
   * @param[in] elbow_to_wrist Length between elbow & wrist (m).
   * @param[in] wrist_to_tip Length between wrist & tip (m).
   */
  virtual bool set_link_length(const double base_to_shoulder, 
                               const double shoulder_to_elbow, 
                               const double elbow_to_wrist, 
                               const double wrist_to_tip);

  /**
   * @brief Get link lengths.
   * @param[out] base_to_shoulder Length between base & shoulder (m).
   * @param[out] shoulder_to_elbow Length between shoulder & elbow (m).
   * @param[out] elbow_to_wrist Length between elbow & wrist (m).
   * @param[out] wrist_to_tip Length between wrist & tip (m).
   */
  virtual void get_link_length(double& base_to_shoulder, 
                               double& shoulder_to_elbow, 
                               double& elbow_to_wrist, 
                               double& wrist_to_tip);
  
 private:
  std::vector<double> stow(const std::vector<double>& xt,
			   const std::vector<std::vector<double> >& Rt);
  bool je(const std::vector<double>& xsw, std::vector<double>& q, 
	  bool positive);
  bool js(const std::vector<double>& xsw, std::vector<double>& q, 
	  bool higher);
  bool jw(const std::vector<std::vector<double> >& Rt, std::vector<double>& q, 
	  bool nearer);
  
  bool je_r(const std::vector<double>& xsw, const std::vector<double>& qr,
	    std::vector<double>& q);
  bool js_r(const std::vector<double>& xsw, const std::vector<double>& qr, 
	    std::vector<double>& q);
  bool jw_r(const std::vector<std::vector<double> >& Rt, const std::vector<double>& qr, 
	    std::vector<double>& q);

  bool invredJ(const std::vector<std::vector<double> >& J, 
	       std::vector<std::vector<double> >& Jinv);
  
  double lbs;
  double lse;
  double lew;
  double lwt;
  //std::vector<double> qb;
  fwd_kin_pa10 fwdkin;
};

#endif
