/**
 * @brief  Inverse kinematics test program
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "inv_kin_pa10.h"
#include <iostream>
#include <cmath>

#define RAD(deg) (deg*M_PI/180.0)
#define DEG(rad) (rad*180.0/M_PI)

int main(int argc, char* argv[])
{
  std::vector<double> jang(7, 0.0);
  std::vector<double> jang_r(7, 0.0);
  std::vector<double> jvel(7, 0.0);
  std::vector<double> jvel_r(7, 0.0);
  std::vector<double> x, x_r;
  std::vector<std::vector<double> > R, R_r;
  std::vector<double> V;
  fwd_kin_pa10 fwd;
  inv_kin_pa10 inv;
  int i, j;
  unsigned char id;
  
  std::cout<<std::fixed;
  //std::cout.precision(3);
  
#if 1
  // Forward Kinematics
  jang[0] = RAD(15.0); // S1
  jang[1] = RAD(30.0); // S2
  jang[2] = 0.0; // S3
  jang[3] = RAD(45.0); // E1
  jang[4] = RAD(15.0); // W1
  jang[5] = RAD(-60.0); // W2
  jang[6] = RAD(180.0); // W3
  if(!fwd.tip_pose(jang, x, R)){
    std::cout<<"kinematics error"<<std::endl;
    return -1;
  }
  std::cout<<" Input joint angle [deg]:";
  for(i=0;i<7;++i)
    std::cout<<" "<<DEG(jang[i]);
  std::cout<<std::endl;
  std::cout<<"Tip position [m]:";
  for(i=0;i<3;++i)
    std::cout<<" "<<x[i];
  std::cout<<std::endl;
  std::cout<<"Tip orientation:"<<std::endl;
  for(i=0;i<3;++i){
    for(j=0;j<3;++j) std::cout<<" "<<R[i][j];
    std::cout<<std::endl;
  }
  
  // Inverse Kinematics
  for(id = 0; id < 8; ++id){
    if(!inv.joint_angle(x, R, jang_r, id)){
      std::cout<<"inverse kinematics error"<<std::endl;
      return -1;
    }
    std::cout<<"Output joint angle [deg]:";
    for(i=0;i<7;++i)
      std::cout<<" "<<DEG(jang_r[i]);
    std::cout<<std::endl;
    
    // verification
    if(!fwd.tip_pose(jang_r, x_r, R_r)){
      std::cout<<"kinematics error"<<std::endl;
      return -1;
    }
    for(i=0;i<3;++i)
      if(fabs(x[i] - x_r[i]) > 1.0e-15){
	std::cout<<(int)id<<": inverse check failed"<<std::endl;
      }
    for(i=0;i<3;++i)
      for(j=0;j<3;++j)
	if(fabs(R[i][j] - R_r[i][j]) > 1.0e-15){
	  std::cout<<(int)id<<": inverse check failed"<<std::endl;
	}
  }
  std::cout<<std::endl;

  if(!inv.joint_angle_opt(x, R, jang, jang_r)){
    std::cout<<"inverse kinematics error"<<std::endl;
    return -1;
  }
  std::cout<<"Optimal joint angle [deg]:";
  for(i=0;i<7;++i)
    std::cout<<" "<<DEG(jang_r[i]);
  std::cout<<std::endl;
  std::cout<<std::endl;
  
#endif
  
#if 1
  std::cout.precision(6);
  
  // Instantaneous kinematics
  jang[0] = RAD(-30.0); // S1
  jang[1] = RAD(20.0); // S2
  jang[2] = 0.0; // S3
  jang[3] = RAD(15.0); // E1
  jang[4] = RAD(-45.0); // W1
  jang[5] = RAD(30.0); // W2
  jang[6] = RAD(15.0); // W3
  
  jvel[0] = RAD(1.0); // S1
  jvel[1] = RAD(2.0); // S2
  jvel[2] = RAD(0.0); // S3
  jvel[3] = RAD(4.0); // E1
  jvel[4] = RAD(5.0); // W1
  jvel[5] = RAD(6.0); // W2
  jvel[6] = RAD(7.0); // W3
  
  if(!fwd.tip_velocity(jang, jvel, V)){
    std::cout<<"kinematics error"<<std::endl;
    return -1;
  }
  std::cout<<"Joint angle [deg]:";
  for(i=0;i<7;++i)
    std::cout<<" "<<DEG(jang[i]);
  std::cout<<std::endl;
  std::cout<<" Input joint velocity [deg/s]:";
  for(i=0;i<7;++i)
    std::cout<<" "<<DEG(jvel[i]);
  std::cout<<std::endl;
  std::cout<<"Tip linear velocity [m/s]:";
  for(i=0;i<3;++i)
    std::cout<<" "<<V[i];
  std::cout<<std::endl;
  std::cout<<"Tip angular velocity [deg/s]:";
  for(i=0;i<3;++i)
    std::cout<<" "<<DEG(V[i+3]);
  std::cout<<std::endl;
  
  if(!inv.joint_velocity(V, jang, jvel_r)){
    std::cout<<"inverse kinematics error"<<std::endl;
    return -1;
  }
  std::cout<<"Output joint velocity [deg/s]:";
  for(i=0;i<7;++i)
    std::cout<<" "<<DEG(jvel_r[i]);
  std::cout<<std::endl;
  
#endif  

  return 0;
}
