// -*- C++ -*-
/*!
 * @file  PA10Manipulator.cpp * @brief PA10 Manipulator Component * $Date$ 
 *
 * $Id$ 
 */
#include "PA10Manipulator.h"

#define PA10_NUM_JOINTS 7

// Module specification
static const char* pa10manipulator_spec[] =
  {
    "implementation_id", "PA10Manipulator",
    "type_name",         "PA10Manipulator",
    "description",       "PA10 Manipulator Component",
    "version",           "1.0",
    "vendor",            "Woo-Keun YOON, Life Robotics Inc.",
    "category",          "pa10manipulator",
    "activity_type",     "SPORADIC",
    "kind",              "PA10 Manipulator",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.mode", "sim",
    "conf.default.tip_len", "0.0",
    "conf.default.sampling", "0.001",
    "conf.default.init_angles", "0.0,0.0,0.0,0.0,0.0,0.0,0.0",
    ""
  };

PA10Manipulator::PA10Manipulator(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    m_JointVelocityIn("JointVelocity", m_JointVelocity),
    m_CurrentJointVelocityOut("CurrentJointVelocity", m_CurrentJointVelocity),
    m_CurrentJointAngleOut("CurrentJointAngle", m_CurrentJointAngle),
    m_DupCurrentJointAngleOut("DupCurrentJointAngle", m_DupCurrentJointAngle),
    m_servo(NULL)

{
  // Registration: InPort/OutPort/Service
  // Set InPort buffers
  registerInPort("JointVelocity", m_JointVelocityIn);

  // Set OutPort buffer
  registerOutPort("CurrentJointAngle", m_CurrentJointAngleOut);
  registerOutPort("Duplicate CurrentJointAngle", m_DupCurrentJointAngleOut);
  registerOutPort("CurrentJointVelocity", m_CurrentJointVelocityOut);
}

PA10Manipulator::~PA10Manipulator()
{
}


RTC::ReturnCode_t PA10Manipulator::onInitialize()
{
  // Bind variables and configuration variable
  bindParameter("mode", m_mode, "sim");
  bindParameter("tip_len", m_tip_len, "0.0");
  bindParameter("sampling", m_sampling, "0.001");
  bindParameter("init_angles", m_init_angles, "0.0,0.0,0.0,0.0,0.0,0.0,0.0");

  return RTC::RTC_OK;
}


RTC::ReturnCode_t PA10Manipulator::onFinalize()
{
  return RTC::RTC_OK;
}

RTC::ReturnCode_t PA10Manipulator::onActivated(RTC::UniqueId ec_id)
{
  if (m_mode.compare("real") == 0) {
    // instantiate the ServoDriverPA10 object
    m_servo = new ServoDriverPA10(0, pa10_info::PA_10A_ARM, m_tip_len);
  }
  else if (m_mode.compare("sim") == 0) {
    // instantiate the ServoSimulatorPA10 object
    m_servo = new ServoSimulatorPA10(m_sampling);
  }
  else {
    std::cout << "Error: mode is wrong. " << std::endl;
    return RTC::RTC_ERROR;
  }

  // initialization of Servo
  if (m_servo == NULL || !(m_servo->initialize())) {
    std::cout << "Error: initialize servo. " << std::endl;
    return RTC::RTC_ERROR;
  }

  if (m_mode.compare("sim") == 0) {
    ServoSimulatorPA10* sim_pa10 = dynamic_cast<ServoSimulatorPA10*>(m_servo);
    sim_pa10->set_sampling_period(m_sampling);
    sim_pa10->set_initial_state(m_init_angles);
  }

  // activation of Servo
  if (!(m_servo->activate())) {
    std::cout << "Error: activate servo. " << std::endl;
    return RTC::RTC_ERROR;
  }

  return RTC::RTC_OK;
}


RTC::ReturnCode_t PA10Manipulator::onDeactivated(RTC::UniqueId ec_id)
{
  // deactivation of Servo
  if (m_servo == NULL || !(m_servo->deactivate())){
    std::cout << "Error: deactivate servo. " << std::endl;
    return RTC::RTC_ERROR;
  }

  // finalization of Servo
  m_servo->finalize();
  delete m_servo;

  return RTC::RTC_OK;
}


RTC::ReturnCode_t PA10Manipulator::onExecute(RTC::UniqueId ec_id)
{
  // read data from InPort
  if (m_servo != NULL && m_JointVelocityIn.isNew()) {
    m_JointVelocityIn.read();

    if (m_JointVelocity.data.length() == PA10_NUM_JOINTS){
      std::vector<double> jointVel(PA10_NUM_JOINTS, 0.0);
      seq_to_vec(m_JointVelocity, jointVel);
      // send joint velocity data
      m_servo->joint_velocity_control(jointVel);
    }
  }


  // get current data
  std::vector<double> joint_displacement(PA10_NUM_JOINTS, 0.0);
  std::vector<double> joint_velocity(PA10_NUM_JOINTS, 0.0);

  if (m_servo != NULL){
    
    if(m_servo->get_current_state(joint_displacement,
				  joint_velocity)) {
      vec_to_seq(joint_displacement, m_CurrentJointAngle);
      vec_to_seq(joint_displacement, m_DupCurrentJointAngle);
      vec_to_seq(joint_velocity, m_CurrentJointVelocity);
      // set data to OutPort
      m_CurrentJointAngleOut.write();
      m_DupCurrentJointAngleOut.write();
      m_CurrentJointVelocityOut.write();
    }
  }

  return RTC::RTC_OK;
}


extern "C"
{
 
  void PA10ManipulatorInit(RTC::Manager* manager)
  {
    RTC::Properties profile(pa10manipulator_spec);
    manager->registerFactory(profile,
                             RTC::Create<PA10Manipulator>,
                             RTC::Delete<PA10Manipulator>);
  }
  
};



