/**
 * @file   ServoDriverBase.h
 * @brief  Servo driver base class definition.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/06
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef SERVO_DRIVER_BASE_H
#define SERVO_DRIVER_BASE_H

#include <vector>

class ServoDriverBase {
 public:
  ServoDriverBase(){}
  virtual ~ServoDriverBase(){}
  
  /**
   * @brief  Initialize servo driver.
   * @return True on success.
   */
  virtual bool initialize() = 0;
  
  /**
   * @brief  Finalize servo driver.
   * @return True on success.
   */
  virtual bool finalize() = 0;
  
  
  /**
   * @brief  Activate servo driver to start control.
   * @return True on success.
   * @note   initialize() must have been called before calling this.
   * @note   Control method has to be invoked periodically 
   *         after activatation succeeded. 
   */
  virtual bool activate() = 0;
  
  /**
   * @brief  Deactivate servo driver to stop control.
   * @return True on success.
   */
  virtual bool deactivate() = 0;
  
  
  /**
   * @brief  Control joint velocity. 
   * @param[in] joint_velocity Commanded joint velocity;  
   *            [rad/s] for revolute joint, [m/s] for prismatic joint. 
   * @return True on success.
   * @note   activate() must have been successfully called before this.
   */
  virtual bool joint_velocity_control(const std::vector<double>& joint_velocity) = 0;
  
  /**
   * @brief  Control joint torque. 
   * @param[in] joint_torque Commanded joint torque; 
   *            [Nm] for revolute joint, [N] for prismatic joint.
   * @return True on success.
   * @note   activate() must have been successfully called before this.
   */
  virtual bool joint_torque_control(const std::vector<double>& joint_torque) = 0;
  
  /**
   * @brief  Get current robot state. (position & velocity)
   * @param[out] joint_displacement Current joint displacement; 
   *             [rad] for revolute joint, [m] for prismatic joint.
   * @paran[out] joint_velocity Current joint velocity; 
   *             [rad/s] for revolute joint, [m/s] for prismatic joint.
   * @return True on success.
   * @note Current state may not be available when servo driver is inactive.
   */
  virtual bool get_current_state(std::vector<double>& joint_displacement, 
				 std::vector<double>& joint_velocity) = 0;
  
};

#endif /* !defined(SERVO_DRIVER_BASE_H) */
