/**
 * @file   ServoDriverPA10.cpp
 * @brief  PA10's servo driver.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 *         Shinji Kurihara
 * @date   2008/10/06
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include <iostream>
#include "ServoDriverPA10.h"

#define PA10_NUM_JOINTS 7

ServoDriverPA10::ServoDriverPA10(unsigned long arm_id, enum pa10_info::TYPE type, double tipLength)
  : m_initialized(false), m_active(false),
    m_curr_angle(PA10_NUM_JOINTS, 0.0), m_curr_speed(PA10_NUM_JOINTS, 0.0),
    m_pa10_info(NULL),    m_pa10_state(NULL),
    m_arm_id(arm_id),     m_arm_type(type),
    m_tipLength(tipLength)
{
}



ServoDriverPA10::~ServoDriverPA10()
{
}



bool ServoDriverPA10::initialize()
{
  if(m_initialized) return false;
  
  // initialization process
  m_pa10_info  = pa10_info_alloc();
  m_pa10_state = pa10_state_alloc();
  if (m_pa10_info == NULL){
    std::cerr << "pa10_info_alloc error." << std::endl;
    return false;
  }
  if (m_pa10_state == NULL){
    std::cerr << "pa10_state_alloc error." << std::endl;
    return false;
  }

  m_initialized = true;
  
  return true;
}



bool ServoDriverPA10::finalize()
{
  if(!m_initialized) return true;
  
  if(m_active) deactivate();
  
  m_initialized = false;
  
  // memory free
  pa10_info_free(m_pa10_info);
  pa10_state_free(m_pa10_state);

  return true;
}



bool ServoDriverPA10::activate()
{
  if(!m_initialized) return false;
  if(m_active) return false;
  
  // initialize state
  for(int i(0); i < PA10_NUM_JOINTS; ++i) m_curr_speed[i] = 0.0;

  m_pa10_info->Id        = m_arm_id;
  m_pa10_info->Type      = m_arm_type;
  m_pa10_info->TipLength = m_tipLength;

  int ret;
  ret = pa10_open(m_pa10_info);
  if(ret != 0) {
    std::cerr << "pa10_open error: " << ret << std::endl;
    return false;
  }

  ret = pa10_start(m_pa10_info);
  if(ret != 0) {
    std::cerr << "pa10_start error: " << ret << std::endl;
    return false;
  }

  m_active = true;
  
  return true;
}



bool ServoDriverPA10::deactivate()
{
  if(!m_active) return true;
  
  m_active = false;
  
  // deactivation
  int ret;
  // sending data stop
  ret = pa10_stop(m_pa10_info);
  // arcnet board close
  ret = pa10_close(m_pa10_info);

  return true;
}



bool ServoDriverPA10::
joint_velocity_control(const std::vector<double>& joint_velocity)
{
  if(!m_active) return false;
  if(joint_velocity.size() != PA10_NUM_JOINTS) return false;
  
  // update state (TODO: check joint&speed limits)
  double jvel[PA10_NUM_JOINTS];
  for(int i(0); i < PA10_NUM_JOINTS; ++i){
    jvel[i] = joint_velocity[i];
  }
  
  int ret;
  ret = pa10_joint_velocity_control(m_pa10_info, jvel);
  if(ret != 0) {
    std::cerr << "joint velocity control error: " << ret << std::endl;
    return false;
  }

  return true;
}



bool ServoDriverPA10::
joint_torque_control(const std::vector<double>& joint_torque)
{
  if(!m_active) return false;
  if(joint_torque.size() != PA10_NUM_JOINTS) return false;

  // update state (TODO: check joint&speed limits)
  double jtorq[PA10_NUM_JOINTS];
  for(int i(0); i < PA10_NUM_JOINTS; ++i){
    jtorq[i] = joint_torque[i];
  }

  int ret;
  ret = pa10_joint_torque_control(m_pa10_info, jtorq);
  if(ret != 0) {
    std::cerr << "joint torque control error: " << ret << std::endl;
    return false;
  }

  return true;

}



bool ServoDriverPA10::get_current_state(std::vector<double>& joint_angle, 
					   std::vector<double>& joint_velocity)
{
  if(!m_active) return false;

  int ret = pa10_get_current_data(m_pa10_info, m_pa10_state);
  if (ret != 0) {
    std::cerr << "get_current_state error: " << ret << std::endl;
    return false;
  }
  
  if (joint_angle.size() != PA10_NUM_JOINTS)    joint_angle.resize(PA10_NUM_JOINTS);
  if (joint_velocity.size() != PA10_NUM_JOINTS) joint_velocity.resize(PA10_NUM_JOINTS);

  int i(0);
  for(i = 0; i < PA10_NUM_JOINTS; i++)
    joint_angle[i] = m_curr_angle[i] = m_pa10_state->CurJoint[i];
  for(i = 0; i < PA10_NUM_JOINTS; i++)
    joint_velocity[i] = m_curr_speed[i] = m_pa10_state->CurJointVelocity[i];
  
  return true;
}
