/**
 * @file   ServoSimulatorPA10.cpp
 * @brief  Simulator of PA10's servo driver.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/06
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "ServoSimulatorPA10.h"


ServoSimulatorPA10::ServoSimulatorPA10(double sampling_period)
  : m_initialized(false), m_active(false),
    m_curr_angle(7, 0.0), m_curr_speed(7, 0.0),
    m_init_angle(7, 0.0), m_period(sampling_period)
{
}

ServoSimulatorPA10::~ServoSimulatorPA10()
{
}

bool ServoSimulatorPA10::initialize()
{
  if(m_initialized) return false;
  
  // initialization process
  
  m_initialized = true;
  
  return true;
}

bool ServoSimulatorPA10::finalize()
{
  if(!m_initialized) return true;
  
  if(m_active) deactivate();
  
  m_initialized = false;
  
  // finalization process
  
  return true;
}

bool ServoSimulatorPA10::activate()
{
  if(!m_initialized) return false;
  if(m_active) return false;
  
  // sanity check
  if(m_period < MINIMUM_SAMPLING_PERIOD) return false;
  if(m_init_angle.size() != 7) return false;
  
  // initialize state
  m_curr_angle = m_init_angle;
  for(int i = 0; i < 7; ++i) m_curr_speed[i] = 0.0;
  m_tick = 0;
  
  m_active = true;
  
  return true;
}

bool ServoSimulatorPA10::deactivate()
{
  if(!m_active) return true;
  
  m_active = false;
  
  // deactivation
  
  return true;
}

bool ServoSimulatorPA10::
joint_velocity_control(const std::vector<double>& joint_velocity)
{
  if(!m_active) return false;
  if(joint_velocity.size() != 7) return false;
  
  // update state (TODO: check joint&speed limits)
  for(int i = 0; i < 7; ++i){
    m_curr_angle[i] += joint_velocity[i]*m_period;
    m_curr_speed[i] = joint_velocity[i];
  }
  ++m_tick;
  
  return true;
}

bool ServoSimulatorPA10::get_current_state(std::vector<double>& joint_angle, 
					   std::vector<double>& joint_velocity)
{
  if(!m_active) return false;
  
  joint_angle = m_curr_angle;
  joint_velocity = m_curr_speed;
  
  return true;
}

bool ServoSimulatorPA10::set_initial_state(const std::vector<double>& joint_angle)
{
  if(m_active) return false;
  if(joint_angle.size() != 7) return false;
  // TODO: check joint limit
  
  m_init_angle = joint_angle;
  
  return true;
}

void ServoSimulatorPA10::get_initial_state(std::vector<double>& joint_angle)
{
  joint_angle = m_init_angle;
}

bool ServoSimulatorPA10::set_sampling_period(double sampling_period)
{
  if(m_active) return false;
  if(sampling_period < MINIMUM_SAMPLING_PERIOD) return false;
  
  m_period = sampling_period;
  
  return true;
}

double ServoSimulatorPA10::get_sampling_period()
{
  return m_period;
}
