// -*- C -*-
/**
 * @file pa10_library.h
 * @brief user interface struct
 * @date $Date: 2008/07/01 $
 * @author Woo-Keun YOON <yoon@liferobotics.jp>
 *
 * Copyright (C) 2008
 *     Life Robitics Inc., Japan
 *     All rights reserved.
 *
 */


#ifndef PA10_LIBRARY_H
#define PA10_LIBRARY_H

/**
 * @brief info of PA10
 * @remarks user inputs initial parameters of PA10
 */
typedef struct{

  unsigned long Id;  // Arm's unique ID (Arm number)
  double TipLength;     // tip length [m]
  enum TYPE {        // type of PA10
    PA_10A_ARM,
    PA10_7C_ARM,
    PA10_6C_ARM,
    PA10_TYPE_ERROR
  } Type; 

} pa10_info;



/**
 * @brief status of PA10
 * @remarks PA10 current condition
 */
typedef struct{

  double CurJoint[7];          // current joint angle [rad]
  double CurTorque[7];         // current torque
  double CurJointVelocity[7];  // current joint velocity [rad/s]
  double CurTipPosition[3];    // current tip position [m]
  double CurTipMatrix[3][3];   // current tip matrix
  unsigned short ServoDriverStatus[8]; // servo driver status
} pa10_state;



/**
 * @brief ARCNET open, iopl(3)
 * @param[in] pa        PA10 arm's property
 * @return 0 on success
 */
int pa10_open(const pa10_info *pa);


/**
 * @brief ARCNET close, iopl(0)
 * @param[in] pa        PA10 arm's property
 * @return 0 on success
 */
int pa10_close(const pa10_info *pa);


/**
 * @brief SERVO start
 * @param[in] pa        PA10 arm's property
 * @return 0 on success
 */
int pa10_start(const pa10_info *pa);


/**
 * @brief SERVO stop
 * @param[in] pa        PA10 arm's property
 * @return 0 on success
 */
int pa10_stop(const pa10_info *pa);


/**
 * @brief Control joint velocity
 * @param[in] pa             PA10 arm's property
 * @param[in] JointVelocity[7]
 * @return 0 on success
 * @note   This is supposed to be invoked periodically.
 */
int pa10_joint_velocity_control(const pa10_info *pa, double JointVelocity[7]);


/**
 * @brief Control joint torque
 * @param[in] pa             PA10 arm's property
 * @param[in] JointTorque[7]
 * @return 0 on success
 * @note   This is supposed to be invoked periodically.
 */
int pa10_joint_torque_control(const pa10_info *pa, double JointTorque[7]);


/**
 * @brief Control tip velocity
 * @param[in] pa             PA10 arm's property
 * @param[in] TipVelocity
 * @return 0 on success
 * @note   This is supposed to be invoked periodically.
 */
int pa10_tip_velocity_control(const pa10_info *pa, double TipVelocity[6]);



/**
 * @brief Current state
 * @param[in] pa             PA10 arm's property
 * @param[out] pa_st         PA10 arm's current state
 * @return 0 on success
 */
int pa10_get_current_data(const pa10_info *pa, pa10_state *pa_st);



/**
 * @brief error recobery
 * @param[in] int axis
 * @param[in] int end_time
 * @return 0 on success
 */
int pa10_joint_limit_error_recovery(const pa10_info *pa, int axis, int end_time);



/**
 * @brief alarm clear
 * @param[in] pa              PA10's property
 * @return 0 on success
 */
int pa10_alarm_clear(const pa10_info *pa);


/**
 * @brief Allocate/Free/Initialize pa10_info structure
 * @return 0 on success
 */
pa10_info *pa10_info_alloc();
int  pa10_info_init(pa10_info *pa);
void pa10_info_free(pa10_info *pa);



/**
 * @brief Allocate/Free/Initialize pa10_state structure
 * @return 0 on success
 */
pa10_state *pa10_state_alloc();
int  pa10_state_init(pa10_state *pa_st);
void pa10_state_free(pa10_state *pa_st);




#endif // PA10_LIBRARY_H
