// -*- C++ -*-
/*!
 * @file   PDControl.cpp 
 * @brief  Generic PD Control RT-Component.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/25
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "PDControl.h"
#include <iostream>
#include "VectorConvert.h"
#include "VecSeqConverter.h"

// Comment next line out in ordinary use
//#define DEBUG_MODE 

// Macros
#define CERR(msg) std::cout<<"[PDControl] ERROR: "<<msg<<std::endl;

#ifdef DEBUG_MODE
#define COUT(msg) std::cout<<"[PDControl] "<<msg<<std::endl
#else
#define COUT(msg)
#endif

#define RETURN_ERROR(msg) do{ CERR(msg); return RTC::RTC_ERROR; }while(0)


// Module specification
static const char* pdcontrol_spec[] = {
  "implementation_id", "PDController",
  "type_name",         "PDController",
  "description",       "Generic PD Controller",
  "version",           "1.0",
  "vendor",            "Masayuki Shimizu",
  "category",          "Controller",
  "activity_type",     "PERIODIC",
  "kind",              "DataFlowComponent",
  "max_instance",      "999",
  "language",          "C++",
  "lang_type",         "compile",
  ""
};

namespace RTC {
  
  PDControl::PDControl(Manager* manager)
    : DataFlowComponentBase(manager),
      m_TargetStateInBuf("TargetState", m_TargetState),
      m_ActualStateInBuf("ActualState", m_ActualState),
      m_ActualSpeedInBuf("ActualSpeed", m_ActualSpeed),
      m_CommandedStateOutBuf("CommandedState", m_CommandedState),
      m_ConfigurationPort("PD Gain Configuration")
  {
    // Registration: InPort/OutPort/Service
    // Set InPort buffers
    registerInPort("Target State", m_TargetStateInBuf);
    registerInPort("Actual State", m_ActualStateInBuf);
    registerInPort("Actual Differential State", m_ActualSpeedInBuf);
    
    // Set OutPort buffer
    registerOutPort("Control Output", m_CommandedStateOutBuf);
    
    // Set service provider to Ports
    m_ConfigurationPort.registerProvider("PDControlSvc0",    // ref name
					 "PDControlService", // interface
					 m_PDControlSvcProvider);
    
    // Set CORBA Service Ports
    registerPort(m_ConfigurationPort);
  }

  PDControl::~PDControl()
  {
    //unregisterPort();
  }

  ReturnCode_t PDControl::onInitialize()
  {
    // bind parameter for configuration
    bindParameter("diag_Pgain", m_conf_diag_Kp, "0.0");
    bindParameter("diag_Dgain", m_conf_diag_Kd, "0.0");
    
    // apply "default" configset
    m_configsets.activateConfigurationSet("default");
    
    return RTC_OK;
  }
  
  ReturnCode_t PDControl::onFinalize()
  {
    // unbind parameter
    return RTC_OK;
  }
  
  ReturnCode_t PDControl::onActivated(UniqueId ec_id)
  {
    COUT("Activated."<<std::endl);
    
    // To avoid data conflict and race condition
    if(ec_id != 0) return RTC_OK;
    
    // get gains from configuration
    if(m_conf_diag_Kp.size() || m_conf_diag_Kd.size()){
      if(!get_gains_from_conf(true))
	RETURN_ERROR("Invalid gain given from configuration.");
    }
    
    // get gains from service port
    if(m_PDControlSvcProvider.is_updated()){
      if(!get_gains_from_svc(true))
	RETURN_ERROR("Invalid gain given from service port.");
    }
    
    // check gains
    std::vector<std::vector<double> > kp;
    std::vector<std::vector<double> > kd;
    m_controller.get_gains(kp, kd);
    if(kp.size() == 0 || kd.size() == 0)
      RETURN_ERROR("No gain given.");
    
#ifdef DEBUG_MODE
    show_gains();
#endif

    // sync 
    m_PDControlSvcProvider.write_gains(kp, kd);
    
    return RTC_OK;
  }
  
  ReturnCode_t PDControl::onDeactivated(UniqueId ec_id)
  {
    COUT("Deactivated."<<std::endl);
    
    if(ec_id != 0) return RTC_OK;
    
    return RTC_OK;
  }
  
  ReturnCode_t PDControl::onExecute(UniqueId ec_id)
  {
    if(ec_id != 0) return RTC_OK;
    
    // get gains from configuration
    if(m_conf_diag_Kp.size() || m_conf_diag_Kd.size()){
      if(!get_gains_from_conf(true))
	RETURN_ERROR("Invalid gain given from configuration.");
#ifdef DEBUG_MODE
      COUT("Configuration updated");
      show_gains();
#endif
    }
    
    // get gains from service port
    if(m_PDControlSvcProvider.is_updated()){
      if(!get_gains_from_svc(true))
	RETURN_ERROR("Invalid gain given from service port.");
#ifdef DEBUG_MODE
      COUT("Service port updated.");
      show_gains();
#endif
    }
    
    if(m_ActualStateInBuf.isNew() && m_ActualSpeedInBuf.isNew()){
      // get current state & speed
      m_ActualState = m_ActualStateInBuf.read();
      m_ActualSpeed = m_ActualSpeedInBuf.read();
      seq_to_vec(m_ActualState, m_actState);
      seq_to_vec(m_ActualSpeed, m_actSpeed);
      
      // get target state if updated
      if(m_TargetStateInBuf.isNew()){
	m_TargetState = m_TargetStateInBuf.read();
	seq_to_vec(m_TargetState, m_tarState);
      }
      
      // control
      if(m_controller.control(m_tarState, m_actState, m_actSpeed, m_output)){
	vec_to_seq(m_output, m_CommandedState);
	m_CommandedStateOutBuf.write(m_CommandedState);
      }
    }
    
    return RTC_OK;
  }
  
  bool PDControl::get_gains_from_conf(bool check_consistency)
  {
    if(!m_controller.set_diag_gains(m_conf_diag_Kp, m_conf_diag_Kd, 
				    check_consistency))
      return false;
    m_conf_diag_Kp.clear();
    m_conf_diag_Kd.clear();
    return true;
  }
  
  bool PDControl::get_gains_from_svc(bool check_consistency)
  {
    std::vector<std::vector<double> > kp;
    std::vector<std::vector<double> > kd;
    if(!m_PDControlSvcProvider.read_gains(kp, kd)) 
      return false;
    if(!m_controller.set_gains(kp, kd, check_consistency)) 
      return false;
    return true;
  }
};


extern "C"
{
  
  void PDControlInit(RTC::Manager* manager)
  {
    RTC::Properties profile(pdcontrol_spec);
    manager->registerFactory(profile,
			     RTC::Create<RTC::PDControl>,
			     RTC::Delete<RTC::PDControl>);
  }
  
};
  
