// -*- C++ -*-
/*!
 * @file   PDControl.h
 * @brief  Generic PD Control RT-Component.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/01
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef RTC_PDCONTROL_H
#define RTC_PDCONTROL_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

#include "PDControlService_impl.h"

#include "pdcontroller.h"

namespace RTC {
  
  class PDControl : public DataFlowComponentBase
  {
  public:
    PDControl(Manager* manager);
    virtual ~PDControl();
    
    virtual ReturnCode_t onInitialize();
    virtual ReturnCode_t onFinalize();
    
    virtual ReturnCode_t onActivated(UniqueId ec_id);
    virtual ReturnCode_t onDeactivated(UniqueId ec_id);
    virtual ReturnCode_t onExecute(UniqueId ec_id);
    
  protected:
    // DataInPort declaration
    TimedDoubleSeq m_TargetState;
    InPort<TimedDoubleSeq, RingBuffer> m_TargetStateInBuf;
    TimedDoubleSeq m_ActualState;
    InPort<TimedDoubleSeq, RingBuffer> m_ActualStateInBuf;
    TimedDoubleSeq m_ActualSpeed;
    InPort<TimedDoubleSeq, RingBuffer> m_ActualSpeedInBuf;
    
    // DataOutPort declaration
    TimedDoubleSeq m_CommandedState;
    OutPort<TimedDoubleSeq, RingBuffer> m_CommandedStateOutBuf;
    
    // CORBA Port declaration
    CorbaPort m_ConfigurationPort;
    
    // Service declaration
    PDControlService_impl m_PDControlSvcProvider;
    
    // controller instance
    pdcontroller m_controller;
    
  private:
    /**
     * @brief Get PD gains from configuration.
     */
    bool get_gains_from_conf(bool check_consistency = true);

    /**
     * @brief Get PD gains from service port.
     */
    bool get_gains_from_svc(bool check_consistency = true);
    
    /**
     * @brief Print PD gains on console for debug use
     */
    inline void show_gains()
    {
      std::vector<std::vector<double> > kp;
      std::vector<std::vector<double> > kd;
      m_controller.get_gains(kp, kd);
      
      unsigned long d = kp.size();
      std::cout<<"P gain matrix: "<<std::endl;
      for(unsigned long i = 0; i < d; ++i){
	std::cout<<"[";
	for(unsigned long j = 0; j < d; ++j) std::cout<<" "<<kp[i][j];
	std::cout<<" ]"<<std::endl;
      }
      std::cout<<std::endl;
      
      std::cout<<"D gain matrix: "<<std::endl;
      for(unsigned long i = 0; i < d; ++i){
	std::cout<<"[";
	for(unsigned long j = 0; j < d; ++j) std::cout<<" "<<kd[i][j];
	std::cout<<" ]"<<std::endl;
      }
      std::cout<<std::endl;
    }
    
    std::vector<double> m_actState;
    std::vector<double> m_actSpeed;
    std::vector<double> m_tarState;
    std::vector<double> m_output;
    
    std::vector<double> m_conf_diag_Kp;
    std::vector<double> m_conf_diag_Kd;
  };
  
};


extern "C"
{
  void PDControlInit(RTC::Manager* manager);
};

#endif // PDCONTROL_H

