// -*-C++-*-
/*!
 * @file   PDControlService_impl.cpp
 * @brief  Service implementation of PD Controller.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/01
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "PDControlService_impl.h"
#include "matrix_util.h"

namespace RTC {
  
  PDControlService_impl::PDControlService_impl()
    : m_updated(false), m_dim(0)
  {
  }
  
  PDControlService_impl::~PDControlService_impl()
  {
  }
  
  CORBA::Boolean PDControlService_impl::set_gains(const DblMatrix& Kp, 
						  const DblMatrix& Kd)
  {
    m_dim = 0;
    unsigned long dim = Kp.length();
    if(!dim) return false;
    if(Kd.length() != dim) return false;
    for(unsigned long i = 0; i < dim; ++i){
      if(Kp[i].length() != dim) return false;
      if(Kd[i].length() != dim) return false;
    }
    if(!is_positive(Kp)) return false;
    if(!is_positive(Kd)) return false;
    
    m_Kp = Kp;
    m_Kd = Kd;
    m_dim = dim;
    m_updated = true;
    
    return true;
  }
  
  CORBA::Boolean PDControlService_impl::get_gains(DblMatrix_out Kp, DblMatrix_out Kd)
  {
    if(!m_dim) return false; 
    
    Kp = new DblMatrix(m_Kp);
    Kd = new DblMatrix(m_Kd);
    
    return true;
  }

  bool PDControlService_impl::is_updated()
  { 
    return m_updated; 
  }

  bool PDControlService_impl::read_gains(std::vector<std::vector<double> >& Kp, 
					 std::vector<std::vector<double> >& Kd)
  {
    m_updated = false;
    
    Kp.resize(m_dim);
    Kd.resize(m_dim);
    for(unsigned long i = 0; i < m_dim; ++i){
      Kp[i].resize(m_dim);
      Kd[i].resize(m_dim);
      for(unsigned long j = 0; j < m_dim; ++j){
	Kp[i][j] = m_Kp[i][j];
	Kd[i][j] = m_Kd[i][j];
      }
    }
    
    return true;
  }
  
  bool PDControlService_impl::write_gains(const std::vector<std::vector<double> >& Kp, 
					  const std::vector<std::vector<double> >& Kd)
  {
    m_dim = 0;
    unsigned long dim = Kp.size();
    if(!dim) return false;
    if(Kd.size() != dim) return false;
    for(unsigned long i = 0; i < dim; ++i){
      if(Kp[i].size() != dim) return false;
      if(Kd[i].size() != dim) return false;
    }
    
    m_Kp.length(dim);
    m_Kd.length(dim);
    for(unsigned long i = 0; i < dim; ++i){
      m_Kp[i].length(dim);
      m_Kd[i].length(dim);
      for(unsigned long j = 0; j < dim; ++j){
	m_Kp[i][j] = Kp[i][j];
	m_Kd[i][j] = Kd[i][j];
      }
    }
    m_dim = dim;
    //m_updated = true;
    
    return true;
  }
  
  bool PDControlService_impl::is_positive(const DblMatrix& x)
  {
    unsigned long d = x.length();
    std::vector<std::vector<double> > m(d);
    for(unsigned long i = 0; i < d; ++i){
      if(x[i].length() != d) return false;
      m[i].resize(d);
      for(unsigned long j = 0; j < d; ++j) m[i][j] = x[i][j];
    }
    
    return matrix_util::is_positive(m);
  }
  
};
