/**
 * @file   matrix_util.cpp
 * @brief  Matrix Computation Utilities
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/09/21
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "matrix_util.h"
#include <cmath>
#include <iostream>

namespace matrix_util {

  bool is_square(const std::vector<std::vector<double> >& x)
  {
    unsigned long dim = x.size();
    if(!dim) return false;
    for(unsigned long i = 0; i < dim; ++i){
      if(x[i].size() != dim) 
	return false;
    }
    return true;
  }
  
  bool is_symmetry(const std::vector<std::vector<double> >& x)
  {
    unsigned long dim = x.size();
    if(!dim) return false;
    
    unsigned long i, j;
    for(i = 0; i < dim; ++i){
      if(x[i].size() != dim) return false;
      for(j = i + 1; j < dim; ++j){
	if(fabs(x[i][j] - x[j][i]) > TOLERANCE)
	  return false;
      }
    }
    
    return true;
  }
  
  bool is_diagonal(const std::vector<std::vector<double> >& x)
  {
    unsigned long dim = x.size();
    if(!dim) return false;
    
    unsigned long i, j;
    for(i = 0; i < dim; ++i){
      if(x[i].size() != dim) return false;
      for(j = 0; j < dim; ++j){
	if(i != j && fabs(x[i][j]) > TOLERANCE)
	  return false;
      }
    }

    return true;
  }
  
  bool is_positive(const std::vector<std::vector<double> >& x)
  {
    if(!is_symmetry(x)) return false;
    
    std::vector<double> u;
    std::vector<std::vector<double> > P;
    if(!jacobi(x, u, P)) return false;
    
    for(unsigned int i = 0; i < u.size(); ++i)
      if(u[i] < TOLERANCE) return false;
    
    return true;
  }
  
  bool eigen_values(const std::vector<std::vector<double> >& x, 
		    std::vector<double>& eigen_values)
  {    
    std::vector<std::vector<double> > P;
    return jacobi(x, eigen_values, P);
  }
  
  bool eigen_vectors(const std::vector<std::vector<double> >& x,
		     std::vector<std::vector<double> >& eigen_vectors)
  { 
    std::vector<double> u;
    return jacobi(x, u, eigen_vectors);
  }
  
  /**
   * @note Re-implementation of jacobi() in "Numerical Recipes in C".
   */
  bool jacobi(const std::vector<std::vector<double> >& x, 
	      std::vector<double>& eigen_values, 
	      std::vector<std::vector<double> >& eigen_vectors)
  {
    if(!is_symmetry(x)) return false;
    
    int n = x.size();
    std::vector<std::vector<double> > a(n);
    std::vector<std::vector<double> > v(n);
    std::vector<double> vp(n), vq(n);
    int i, j;
    
    // initialize
    for(i = 0; i < n; ++i){ 
      a[i].resize(n);
      v[i].resize(n);
      for(j = 0; j < n; ++j){
	a[i][j] = x[i][j];
	v[i][j] = 0.0;
      }
      v[i][i] = 1.0;
    }
    
    int ip, iq;
    double tresh, theta, t, sm, s, h, c;
    for(i = 1; i <= 50; ++i){
      // sum of non-diagonal elements
      sm = 0.0;
      for(ip = 0; ip < n-1; ++ip){
	for(iq = ip+1; iq < n; ++iq)
	  sm += fabs(a[ip][iq]);
      }
      if(sm < TOLERANCE) break;
      
      //std::cout<<"sum = "<<sm<<std::flush<<std::endl;
      
      // sweep
      if(i < 4) tresh = 0.2*sm/(double)(n*n);
      else tresh = 0.0;
      for(ip = 0; ip < n-1; ++ip){
	for(iq = ip+1; iq < n; ++iq){
	  if(i > 4 && 
	     fabs(a[ip][ip]) < TOLERANCE && fabs(a[iq][iq]) < TOLERANCE)
	    a[ip][iq] = 0.0; // no need to rotate
	  else if(fabs(a[ip][iq]) > tresh){
	    // t
	    h = a[iq][iq] - a[ip][ip];
	    if(fabs(a[ip][iq]) < TOLERANCE)
	      t = h/a[ip][iq];
	    else{
	      theta = 0.5*h/a[ip][iq];
	      t = fabs(theta) + sqrt(1.0 + theta*theta);
	      if(theta < 0.0) t = -t;
	    }
	    // sin cos tau
	    c = 1.0/sqrt(1.0+t*t);
	    s = t*c;
	    //std::cout<<"sin = "<<s<<" cos = "<<c<<std::flush<<std::endl;
	    // rotation
	    h = t*a[ip][iq];
	    for(j = 0; j < n; ++j){
	      if(j != ip && j != iq){
		a[j][ip] = c*a[j][ip] - s*a[j][iq];
		a[j][iq] = c*a[j][iq] + s*a[j][ip];
	      }
	    }
	    a[ip][ip] -= h;
	    a[iq][iq] += h;
	    a[ip][iq] = 0.0;
	    
	    for(j = 0; j < n; ++j){
	      vp[j] = c*v[j][ip] - s*v[j][iq];
	      vq[j] = s*v[j][ip] + c*v[j][iq];
	    }
	    for(j = 0; j < n; ++j){
	      v[j][ip] = vp[j];
	      v[j][iq] = vq[j];
	    }
	    
	  } // else if
	} // iq
      } // ip
    }
    
    if(i >= 50) return false;
    
    eigen_values.resize(n);
    eigen_vectors.resize(n);
    for(i = 0; i < n; i++){ 
      eigen_values[i] = a[i][i];
      eigen_vectors[i] = v[i];
    }
    
    return true;
  }
  
};
