/**
 * @file   matrix_util.h
 * @brief  Matrix Computation Utilities
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/02
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef MATRIX_UTIL_H
#define MATRIX_UTIL_H

#include <vector>

#undef TOLERANCE
#define TOLERANCE (1.0e-12)

namespace matrix_util {

  /**
   * @brief Check if a matrix is square.
   * @param[in] x Matrix to be checked.
   * @return True if it is square, otherwise false.
   */ 
  bool is_square(const std::vector<std::vector<double> >& x);

  /**
   * @brief Check if a matrix is symmetric.
   * @param[in] x Matrix to be checked.
   * @return True if it is symmetric, otherwise false.
   */ 
  bool is_symmetry(const std::vector<std::vector<double> >& x);

  /**
   * @brief Check if a matrix is diagonal.
   * @param[in] x Matrix to be checked.
   * @return True if it is diagonal, otherwise false.
   */ 
  bool is_diagonal(const std::vector<std::vector<double> >& x);

  /**
   * @brief Check if a matrix is symmetric positive definite.
   * @param[in] x Matrix to be checked.
   * @return True if it is positive definite, otherwise false.
   */ 
  bool is_positive(const std::vector<std::vector<double> >& x);
  

  /**
   * @brief Compute eigenvalues of a matrix.
   * @param[in] x Matrix.
   * @param[out] eigen_values Eigenvalues of the matrix.
   * @return True on success, otherwise false.
   */
  bool eigen_values(const std::vector<std::vector<double> >& x, 
		    std::vector<double>& eigen_values);

  /**
   * @brief Compute eigenvectors of a matrix.
   * @param[in] x Matrix.
   * @param[out] eigen_vectors Set of eigenvectors of the matrix.
   *             Each column is an eigenvector.
   * @return True on success, otherwise false.
   */
  bool eigen_vectors(const std::vector<std::vector<double> >& x,
		     std::vector<std::vector<double> >& eigen_vectors);
  
  /**
   * @brief Compute eigenvalues and eigenvectors based on Jacobi method.
   * @param[in] x Matrix.
   * @param[out] eigen_values Eigenvalues of the matrix.
   * @param[out] eigen_vectors Set of eigenvectors of the matrix.
   *             Each column is an eigenvector.
   * @return True on success, otherwise false.
   * @note The input matrix has to be symmetric. 
   */
  bool jacobi(const std::vector<std::vector<double> >& x, 
	      std::vector<double>& eigen_values, 
	      std::vector<std::vector<double> >& eigen_vectors);
};

#endif
