/**
 * @file   pdcontroller.cpp
 * @brief  Generic PD controller.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/02
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#include "pdcontroller.h"
#include "matrix_util.h"


//#define DEBUG_MODE

#ifdef DEBUG_MODE
#include <iostream>
#define COUT(msg) std::cout<<"[pdcontroller] "<<msg<<std::endl
#else
#define COUT(msg)
#endif

#define RETURN_FALSE(msg) do{ COUT(msg); return false; }while(0)


pdcontroller::pdcontroller()
  : m_period(0.0), m_dim(0)
{
}

pdcontroller::pdcontroller(const pdcontroller& x)
  : m_Kp(x.m_Kp), m_Kd(x.m_Kd), 
    m_period(x.m_period), m_dim(x.m_dim),
    m_prev_state(x.m_prev_state)
{
}

pdcontroller::~pdcontroller()
{
}


bool pdcontroller::control(const std::vector<double>& target_state, 
			   const std::vector<double>& actual_state,
			   const std::vector<double>& actual_speed, 
			   std::vector<double>& output)
{
  // sanity check
  if(!m_dim) 
    RETURN_FALSE("No gain specified");
  if(target_state.size() != m_dim) 
    RETURN_FALSE("Vector size error: target_state");
  if(actual_state.size() != m_dim)
    RETURN_FALSE("Vector size error: actual_state");
  if(actual_speed.size() != m_dim)
    RETURN_FALSE("Vector size error: actual_speed");
  
  // compute control output
  unsigned long i,j;
  output.resize(m_dim);
  for(i = 0; i < m_dim; ++i){
    output[i] = 0.0;
    for(j = 0; j < m_dim; ++j){
      output[i] += m_Kp[i][j]*(target_state[j] - actual_state[j]); 
      output[i] -= m_Kd[i][j]*actual_speed[j];
    }
  }
  
  return true;
}

bool pdcontroller::periodic_control(const std::vector<double>& target_state, 
				    const std::vector<double>& actual_state,
				    std::vector<double>& output)
{
  // sanity check
  if(!m_dim) RETURN_FALSE("No gain specified");
  if(target_state.size() != m_dim) 
    RETURN_FALSE("Vector size error: target_state");
  if(actual_state.size() != m_dim) 
    RETURN_FALSE("Vector size error: actual_state");
  if(m_period < MIN_PERIOD) 
    RETURN_FALSE("Invalid sampling period");
  
  // compute control output
  unsigned long i,j;
  if(m_prev_state.size() == m_dim){
    output.resize(m_dim);
    for(i = 0; i < m_dim; ++i){
      output[i] = 0.0;
      for(j = 0; j < m_dim; ++j){
	output[i] += m_Kp[i][j]*(target_state[j] - actual_state[j]); 
	output[i] -= m_Kd[i][j]*(actual_state[j] - m_prev_state[j])/m_period;
      }
    }
  }else if(m_prev_state.size() == 0){
    output.resize(m_dim);
    for(i = 0; i < m_dim; ++i){
      output[i] = 0.0;
      for(j = 0; j < m_dim; ++j){
	output[i] += m_Kp[i][j]*(target_state[j] - actual_state[j]); 
	output[i] -= m_Kd[i][j]*actual_state[j]/m_period;
      }
    }
  }else{
    COUT("Unexpected error");
    return false;
  }
  
  // update state
  m_prev_state = actual_state;
  
  return true;
}

void pdcontroller::clear()
{
  m_prev_state.clear();
}

bool pdcontroller::set_gains(const std::vector<std::vector<double> >& Kp,
			     const std::vector<std::vector<double> >& Kd, 
			     bool check_consistency)
{
  m_dim = 0;
  
  // sanity check
  unsigned long dim = Kp.size();
  if(!dim) RETURN_FALSE("Kp is nil");
  if(Kd.size() != dim) RETURN_FALSE("Kd's dimension must be same as Kp's"); 
  for(unsigned long i = 0; i < dim; ++i){
    if(Kp[i].size() != dim) RETURN_FALSE("Kp is not square");
    if(Kd[i].size() != dim) RETURN_FALSE("Kd is not square");
  }
  
  if(check_consistency){
    if(!matrix_util::is_symmetry(Kp) || !matrix_util::is_positive(Kp))
      RETURN_FALSE("Kp is not symmetric positive definite");
    if(!matrix_util::is_symmetry(Kd) || !matrix_util::is_positive(Kd)) 
      RETURN_FALSE("Kd is not symmetric positive definite");
  }
  
  // set gains
  m_Kp = Kp;
  m_Kd = Kd;
  m_dim = dim;
  
  return true;
}

bool pdcontroller::set_diag_gains(const std::vector<double>& diag_Kp, 
				  const std::vector<double>& diag_Kd, 
				  bool check_consistency)
{
  m_dim = 0;
  
  // sanity check
  unsigned long dim = diag_Kp.size();
  if(!dim) RETURN_FALSE("Kp is nil");
  if(diag_Kd.size() != dim) 
    RETURN_FALSE("Kd's dimension must be same as Kp's");
  
  if(check_consistency){
    for(unsigned long i = 0; i < dim; ++i){
      if(diag_Kp[i] <= TOLERANCE) 
	RETURN_FALSE("Kp is not positive definite");
      if(diag_Kd[i] <= TOLERANCE)
	RETURN_FALSE("Kd is not positive definite");
    }
  }
  
  // set gains
  m_Kp.resize(dim);
  m_Kd.resize(dim);
  for(unsigned long i = 0; i < dim; ++i){
    m_Kp[i].resize(dim);
    m_Kd[i].resize(dim);
    for(unsigned long j = 0; j < dim; ++j){
      m_Kp[i][j] = 0.0;
      m_Kd[i][j] = 0.0;
    }
    m_Kp[i][i] = diag_Kp[i];
    m_Kd[i][i] = diag_Kd[i];
  }
  m_dim = dim;
  
  return true;
}

bool pdcontroller::set_sampling_period(double period)
{
  if(period < MIN_PERIOD)
    RETURN_FALSE("Too short sampling period");
  m_period = period;
  return true;
}

void pdcontroller::get_gains(std::vector<std::vector<double> >& Kp,
			     std::vector<std::vector<double> >& Kd)
{
  Kp.resize(m_dim);
  Kd.resize(m_dim);
  for(unsigned long i = 0; i < m_dim; ++i){
    Kp[i].resize(m_dim);
    Kd[i].resize(m_dim);
    for(unsigned long j = 0; j < m_dim; ++j){
      Kp[i][j] = m_Kp[i][j];
      Kd[i][j] = m_Kd[i][j];
    }
  }
}

double pdcontroller::get_sampling_period()
{
  return m_period;
}
