/**
 * @file   pdcontroller.h
 * @brief  Generic PD controller.
 * @author Masayuki Shimizu <tmsimiz@ipc.shizuoka.ac.jp>
 * @date   2008/10/02
 * @note   Copyright (C) 2008 Masayuki Shimizu. All Rights Reserved.
 */

#ifndef PDCONTROLLER_H
#define PDCONTROLLER_H

#include <vector>
#include "matrix_util.h"

/* Minimal sampling period [s] */
#define MIN_PERIOD (1.0e-6)

class pdcontroller {
 public:
  pdcontroller();
  pdcontroller(const pdcontroller& x);
  virtual ~pdcontroller();
  
  /**
   * @brief Compute control output.
   * @param[in] target_state Target state.
   * @param[in] actual_state Current state.
   * @param[in] actual_speed Time derivative of current state.
   * @param[out] output Control output to be passed to robot/system.
   * @return - True on success. 
   *         - False, if the dimension of an input vector is different from 
   *           the dimension of PD gains, false is returned.
   *         - False, if PD gains are not correctly specified.
   */
  virtual bool control(const std::vector<double>& target_state, 
		       const std::vector<double>& actual_state,
		       const std::vector<double>& actual_speed, 
		       std::vector<double>& output);
  
  /**
   * @brief Compute control output. (periodic version)
   * @param[in] target_state Target state.
   * @param[in] actual_state Current state.
   * @param[out] output Control output to be passed to robot/system.
   * @return - True on success. 
   *         - False, if the dimension of an input vector is different from 
   *           the dimension of PD gains. 
   *         - False, if PD gains are not correctly specified.
   *         - False, if sampling period is not correctly specified.
   * @note Calling this function should be exactly periodic.
   */
  virtual bool periodic_control(const std::vector<double>& target_state, 
				const std::vector<double>& actual_state,
				std::vector<double>& output);
  
  /**
   * @brief Clear internal variables.
   * @note  This has to be called before starting periodic control.  
   */
  virtual void clear();
  
  
  /**
   * @brief Set PD gain matrices. 
   * @param[in] Kp Proportional gain matrix. 
   * @param[in] Kd Derivative gain matrix. 
   * @return - True on success.
   *         - False, if the dimensions of the Kp and Kd are not same. 
   *         - False, if the matrices are not symmetric positive definite.
   */
  virtual bool set_gains(const std::vector<std::vector<double> >& Kp,
			 const std::vector<std::vector<double> >& Kd, 
			 bool check_consistency = true);
  
  /**
   * @brief Set diagonal PD gain matrices. 
   * @param[in] diag_Kp Diagonal elements of proportional gain matrix.
   * @param[in] diag_Kd Diagonal elements of derivative gain matrix.
   * @return - True on success.
   *         - False, if the dimensions of the Kp and Kd are not same. 
   *         - False, if the matrices are not symmetric positive definite.
   */
  virtual bool set_diag_gains(const std::vector<double>& diag_Kp, 
			      const std::vector<double>& diag_Kd,
			      bool check_consistency = true);
  
  /**
   * @brief Set sampling period. 
   * @param[in] period Sampling period [s].
   * @return - True on success.
   *         - False, if the given value is too small. (< ~1.0e-6)
   * @note This is needless if you do not use the periodic control function.
   */
  virtual bool set_sampling_period(double period);
  
  /**
   * @brief Get PD gain matrices.
   * @param[out] Kp Proportional gain matrices.
   * @param[out] Kd Derivative gain matrices.
   */
  virtual void get_gains(std::vector<std::vector<double> >& Kp,
			 std::vector<std::vector<double> >& Kd);

  /**
   * @brief Get sampling period.
   * @return Sampling period [s].
   */
  virtual double get_sampling_period();
  
 private:
  // properties
  std::vector<std::vector<double> > m_Kp;
  std::vector<std::vector<double> > m_Kd;
  double m_period;
  unsigned long m_dim;
  
  // internal variables
  std::vector<double> m_prev_state;
};


#endif /* !defined(PDCONTROLLER_H) */
